"""
The current fastest way to draw text with arcade.

This example improves on the other two text-drawing examples
by using pyglet's batch functionality.

Although pyglet's batches do not support non-drawing features like
Arcade's SpriteList, they offer similar benefits for drawing. Adding
arcade.Text objects to a batch allows drawing thousands of them with
almost the same cost as drawing a single one directly.

If Python and Arcade are installed, this example can be run from the command line with:
python -m arcade.examples.drawing_text_objects_batch
"""
import arcade
from pyglet.graphics import Batch

# Load fonts bundled with Arcade such as the Kenney fonts
arcade.resources.load_kenney_fonts()
arcade.resources.load_liberation_fonts()

WINDOW_WIDTH = 1280  # Window width in pixels
WINDOW_HEIGHT = 800  # Window height in pixels
WINDOW_TITLE = "Drawing Text Example"  # Window title
DEFAULT_LINE_HEIGHT = 45  # Line height to use in pixels
DEFAULT_FONT_SIZE = 20  # Default font size in points


class GameView(arcade.View):
    """
    Main application class.
    """

    def __init__(self):
        super().__init__()

        self.background_color = arcade.color.BEIGE
        self.text_angle = 0
        self.time_elapsed = 0.0

        self.batch = Batch()

        # Add the screen title
        start_x = 0
        start_y = WINDOW_HEIGHT - DEFAULT_LINE_HEIGHT * 1.5
        self.title = arcade.Text(
            "Text Drawing Examples",
            start_x,
            start_y,
            arcade.color.BLACK,
            DEFAULT_FONT_SIZE * 2,
            width=WINDOW_WIDTH,
            align="center",
            batch=self.batch,
        )

        # start_x and start_y make the start point for the text. We draw a dot to make it
        # easy too see the text in relation to its start x and y.
        start_x = 10
        start_y = WINDOW_HEIGHT - DEFAULT_LINE_HEIGHT * 3
        self.fonts = arcade.Text(
            "Fonts:",
            start_x,
            start_y,
            arcade.color.FRENCH_WINE,
            DEFAULT_FONT_SIZE, bold=True,
            batch=self.batch,
        )

        # Move the y value down to create another line of text
        start_y -= DEFAULT_LINE_HEIGHT
        self.font_default = arcade.Text(
            "Default Font (Arial)",
            start_x,
            start_y,
            arcade.color.BLACK,
            DEFAULT_FONT_SIZE,
            batch=self.batch,
        )

        # Show some built-in fonts
        start_y -= DEFAULT_LINE_HEIGHT
        self.font_kenney_blocks = arcade.Text(
            "Kenney Blocks Font",
            start_x,
            start_y,
            arcade.color.BLACK,
            DEFAULT_FONT_SIZE,
            font_name="Kenney Blocks",
            batch=self.batch,
        )

        start_y -= DEFAULT_LINE_HEIGHT
        self.font_kenney_future = arcade.Text(
            "Kenney Future Font",
            start_x,
            start_y,
            arcade.color.BLACK,
            DEFAULT_FONT_SIZE,
            font_name="Kenney Future",
            batch=self.batch,
        )

        start_y -= DEFAULT_LINE_HEIGHT
        self.font_kenney_high = arcade.Text(
            "Kenney High Font",
            start_x,
            start_y,
            arcade.color.BLACK,
            DEFAULT_FONT_SIZE,
            font_name="Kenney High",
            batch=self.batch,
        )

        start_y -= DEFAULT_LINE_HEIGHT
        self.font_kenney_high_square = arcade.Text(
            "Kenney High Square Font",
            start_x,
            start_y,
            arcade.color.BLACK,
            DEFAULT_FONT_SIZE,
            font_name="Kenney High Square",
            batch=self.batch,
        )

        start_y -= DEFAULT_LINE_HEIGHT
        self.font_kenney_mini_square = arcade.Text(
            "Kenney Mini Square Font",
            start_x, start_y,
            arcade.color.BLACK,
            DEFAULT_FONT_SIZE,
            font_name="Kenney Mini Square",
            batch=self.batch,
        )

        start_y -= DEFAULT_LINE_HEIGHT
        self.font_kenney_pixel = arcade.Text(
            "Kenney Pixel Font",
            start_x, start_y,
            arcade.color.BLACK,
            DEFAULT_FONT_SIZE,
            font_name="Kenney Pixel",
            batch=self.batch,
        )

        start_y -= DEFAULT_LINE_HEIGHT
        self.font_kenney_pixel_square = arcade.Text(
            "Kenney Pixel Square Font",
            start_x, start_y,
            arcade.color.BLACK,
            DEFAULT_FONT_SIZE,
            font_name="Kenney Pixel Square",
            batch=self.batch,
        )

        start_y -= DEFAULT_LINE_HEIGHT
        self.font_kenney_rocket = arcade.Text(
            "Kenney Rocket Font",
            start_x, start_y,
            arcade.color.BLACK,
            DEFAULT_FONT_SIZE,
            font_name="Kenney Rocket",
            batch=self.batch,
        )

        start_y -= DEFAULT_LINE_HEIGHT
        self.font_kenney_rocket_square = arcade.Text(
            "Kenney Rocket Square Font",
            start_x, start_y,
            arcade.color.BLACK,
            DEFAULT_FONT_SIZE,
            font_name="Kenney Rocket Square",
            batch=self.batch,
        )

        start_y -= DEFAULT_LINE_HEIGHT
        # When trying to use system fonts, it can be risky to specify
        # only a single font because someone else's computer might not
        # have it installed. This is especially true if they run a
        # different operating system. For example, if you are on Windows
        # and a friend has a mac or Linux, they might not have the same
        # fonts. Your game could look different or broken on their computer.
        # One way around that is to provide multiple options for draw_text
        # to try. It will use the first one it finds, and use Arial as a
        # default if it can't find any of them.
        # In the example below, draw_text is given a tuple of names for very
        # similar fonts, each of which is common on a different major
        # operating systems.
        self.font_times_new_roman = arcade.Text(
            "Times New Roman (Or closest match on system)",
            start_x, start_y,
            arcade.color.BLACK,
            DEFAULT_FONT_SIZE,
            font_name=(
                "Times New Roman",  # Comes with Windows
                "Times",  # MacOS may sometimes have this variant
                # Common on Linux systems + we ship it with Arcade
                "Liberation Serif"
            ),
            batch=self.batch,
        )

        start_y -= DEFAULT_LINE_HEIGHT
        self.multi_line_breaks = arcade.Text(
            "Multi-Line\ntext using\n\\n characters.",
            start_x, start_y,
            arcade.color.BLACK,
            DEFAULT_FONT_SIZE / 2,
            multiline=True,
            width=300,
            batch=self.batch,
        )

        start_y -= DEFAULT_LINE_HEIGHT * 1.5
        self.multiline_wrap = arcade.Text(
            "Wrapping really long text automatically to a new line. "
            "The quick brown fox jumped over the lazy dogs.",
            start_x,
            start_y,
            arcade.color.BLACK,
            DEFAULT_FONT_SIZE / 2,
            multiline=True,
            width=300,
            batch=self.batch,
        )

        # --- Column 2 ---
        start_x = 750
        start_y = WINDOW_HEIGHT - DEFAULT_LINE_HEIGHT * 3
        self.text_positioning = arcade.Text(
            "Text Positioning:",
            start_x,
            start_y,
            arcade.color.FRENCH_WINE,
            DEFAULT_FONT_SIZE,
            bold=True,
            batch=self.batch,
        )

        # start_x and start_y make the start point for the text.
        # We draw a dot to make it easy too see the text in relation to
        # its start x and y.
        start_y -= DEFAULT_LINE_HEIGHT

        self.default_baseline_left = arcade.Text(
            "Default of 'baseline' and 'Left'",
            start_x,
            start_y,
            arcade.color.BLACK,
            DEFAULT_FONT_SIZE,
            batch=self.batch,
        )

        start_y -= DEFAULT_LINE_HEIGHT
        self.bottom_left = arcade.Text(
            "'bottom' and 'left'",
            start_x,
            start_y,
            arcade.color.BLACK,
            DEFAULT_FONT_SIZE,
            anchor_x="left",
            anchor_y="bottom",
            batch=self.batch,
        )

        start_y -= DEFAULT_LINE_HEIGHT
        self.top_left = arcade.Text(
            "'top' and 'left'",
            start_x, start_y,
            arcade.color.BLACK,
            DEFAULT_FONT_SIZE,
            anchor_x="left",
            anchor_y="top",
            batch=self.batch,
        )

        start_y -= DEFAULT_LINE_HEIGHT * 2
        self.basline_center = arcade.Text(
            "'baseline' and 'center'",
            start_x, start_y,
            arcade.color.BLACK,
            DEFAULT_FONT_SIZE,
            anchor_x="center",
            anchor_y="baseline",
            batch=self.batch,
        )

        start_y -= DEFAULT_LINE_HEIGHT
        self.baseline_right = arcade.Text(
            "'baseline' and 'right'",
            start_x,
            start_y,
            arcade.color.BLACK,
            DEFAULT_FONT_SIZE,
            anchor_x="right",
            anchor_y="baseline",
            batch=self.batch,
        )

        start_y -= DEFAULT_LINE_HEIGHT
        self.center_center = arcade.Text(
            "'center' and 'center'",
            start_x,
            start_y,
            arcade.color.BLACK,
            DEFAULT_FONT_SIZE,
            anchor_x="center",
            anchor_y="center",
            batch=self.batch,
        )

        start_y -= DEFAULT_LINE_HEIGHT * 4
        # start_x = 0
        # start_y = 0
        self.rotating_text = arcade.Text(
            "Rotating Text",
            start_x, start_y,
            arcade.color.BLACK,
            DEFAULT_FONT_SIZE,
            anchor_x="center",
            anchor_y="center",
            rotation=0,
            batch=self.batch,
        )

    def on_update(self, delta_time):
        self.text_angle += 1
        self.time_elapsed += delta_time
        self.rotating_text.rotation = self.text_angle

    def on_draw(self):
        """
        Render the screen.
        """
        # This command should happen before we start drawing. It will clear
        # the screen to the background color, and erase what we drew last frame.
        self.clear()

        # Draw all the text objects
        self.batch.draw()

        arcade.draw_point(
            self.default_baseline_left.x,
            self.default_baseline_left.y,
            arcade.color.BARN_RED,
            5,
        )

        arcade.draw_point(
            self.bottom_left.x,
            self.bottom_left.y,
            arcade.color.BARN_RED,
            5,
        )

        arcade.draw_point(
            self.top_left.x,
            self.top_left.y,
            arcade.color.BARN_RED,
            5,
        )

        arcade.draw_point(
            self.basline_center.x,
            self.basline_center.y,
            arcade.color.BARN_RED,
            5
        )

        arcade.draw_point(
            self.baseline_right.x,
            self.baseline_right.y,
            arcade.color.BARN_RED,
            5,
        )

        arcade.draw_point(
            self.center_center.x,
            self.center_center.y,
            arcade.color.BARN_RED,
            5,
        )

        arcade.draw_point(
            self.rotating_text.x,
            self.rotating_text.y,
            arcade.color.BARN_RED,
            5,
        )

    def on_key_press(self, symbol: int, modifiers: int):
        """ Handle key press events """
        if symbol == arcade.key.ESCAPE:
            self.window.close()


def main():
    """ Main function """
    # Create a window class. This is what actually shows up on screen
    window = arcade.Window(WINDOW_WIDTH, WINDOW_HEIGHT, WINDOW_TITLE)

    # Create and setup the GameView
    game = GameView()

    # Show GameView on screen
    window.show_view(game)

    # Start the arcade game loop
    arcade.run()


if __name__ == "__main__":
    main()



from arcade import gl
from arcade.types import RGBOrA255

from .helpers import _generic_draw_line_strip
from .polygon import draw_polygon_outline


def draw_triangle_filled(
    x1: float, y1: float, x2: float, y2: float, x3: float, y3: float, color: RGBOrA255
) -> None:
    """
    Draw a filled in triangle.

    Args:
        x1: x value of first coordinate.
        y1: y value of first coordinate.
        x2: x value of second coordinate.
        y2: y value of second coordinate.
        x3: x value of third coordinate.
        y3: y value of third coordinate.
        color: Color of the triangle as an RGBA :py:class:`tuple` or
            :py:class:`.Color` instance.
    """
    point_list = (
        (x1, y1),
        (x2, y2),
        (x3, y3),
    )
    _generic_draw_line_strip(point_list, color, gl.TRIANGLES)


def draw_triangle_outline(
    x1: float,
    y1: float,
    x2: float,
    y2: float,
    x3: float,
    y3: float,
    color: RGBOrA255,
    border_width: float = 1,
) -> None:
    """
    Draw a the outline of a triangle.

    Args:
        x1: x value of first coordinate.
        y1: y value of first coordinate.
        x2: x value of second coordinate.
        y2: y value of second coordinate.
        x3: x value of third coordinate.
        y3: y value of third coordinate.
        color: RGBOrA255 of triangle as an RGBA
            :py:class:`tuple` or :py:class:`.Color` instance.
        border_width: Width of the border in pixels. Defaults to 1.
    """
    point_list = (
        (x1, y1),
        (x2, y2),
        (x3, y3),
    )
    draw_polygon_outline(point_list, color, border_width)



"""
This module pre-defines colors from FlatUI.
https://materialui.co/flatuicolors/
"""

from arcade.color import BLACK, WHITE, Color

WHITE = WHITE
GREEN_TURQUOISE = Color(26, 188, 156, 255)
GREEN_GREEN_SEA = Color(22, 160, 133, 255)
GREEN_NEPHRITIS = Color(39, 174, 96, 255)
GREEN_EMERALD = Color(46, 204, 113, 255)
BLUE_PETER_RIVER = Color(52, 152, 219, 255)
BLUE_BELIZE_HOLE = Color(41, 128, 185, 255)
PURPLE_AMETHYST = Color(155, 89, 182, 255)
PURPLE_WISTERIA = Color(142, 68, 173, 255)
DARK_BLUE_WET_ASPHALT = Color(52, 73, 94, 255)
DARK_BLUE_MIDNIGHT_BLUE = Color(44, 62, 80, 255)
YELLOW_SUN_FLOWER = Color(241, 196, 15, 255)
YELLOW_ORANGE = Color(243, 156, 18, 255)
ORANGE_CARROT = Color(230, 126, 34, 255)
ORANGE_PUMPKIN = Color(211, 84, 0, 255)
RED_ALIZARIN = Color(231, 76, 60, 255)
RED_POMEGRANATE = Color(192, 57, 43, 255)
WHITE_CLOUDS = Color(236, 240, 241, 255)
WHITE_SILVER = Color(189, 195, 199, 255)
GRAY_CONCRETE = Color(149, 165, 166, 255)
GRAY_ASBESTOS = Color(127, 140, 141, 255)
BLACK = BLACK

__all__ = [
    "BLACK",
    "WHITE",
    "GREEN_TURQUOISE",
    "GREEN_GREEN_SEA",
    "GREEN_NEPHRITIS",
    "GREEN_EMERALD",
    "BLUE_PETER_RIVER",
    "BLUE_BELIZE_HOLE",
    "PURPLE_AMETHYST",
    "PURPLE_WISTERIA",
    "DARK_BLUE_WET_ASPHALT",
    "DARK_BLUE_MIDNIGHT_BLUE",
    "YELLOW_SUN_FLOWER",
    "YELLOW_ORANGE",
    "ORANGE_CARROT",
    "ORANGE_PUMPKIN",
    "RED_ALIZARIN",
    "RED_POMEGRANATE",
    "WHITE_CLOUDS",
    "WHITE_SILVER",
    "GRAY_CONCRETE",
    "GRAY_ASBESTOS",
]



"""
Simple video recorder with cv2.
Records Arcade screen to a video file, with minimal
I/O operations and minimal memory usage.

We read pixels directly from OpenGL framebuffer, then into
a NumPy array that we can hand to OpenCV for the video writing.

The other ways of doing this with built-in arcade.get_image()
function would require either a lot of re-processing of images
on CPU for no real reason, saving/reloading each frame as an image file,
or having rampant memory usage problems.

This attempts to have optimal performance while not increasing memory usage
by having as few steps in between the pixel data in VRAM and the final video
file as possible.

Dependencies:
    pip install opencv-python numpy
"""

import cv2  # type: ignore
import numpy  # type: ignore
import pyglet.gl as gl

import arcade

SCREEN_WIDTH = 800
SCREEN_HEIGHT = 600
SCREEN_TITLE = "Starting Template"


class VideoRecorderCV2(arcade.Window):
    def __init__(self, width, height, title):
        super().__init__(width, height, title)
        arcade.set_background_color(arcade.color.BLACK)

        # The video files produced by this are un-compressed and will be massive.
        # Think 1GB+ for even like 15 seconds+ of 1920x1080. If you want compressed
        # video, you can replace the below line with something like this to use H.264
        # encoding(this probably needs ffmpeg available):
        #
        # fourcc = cv2.VideoWriter_fourcc(*'mp4v')
        # self.video = cv2.VideoWriter("my_video.mp4", fourcc, 60, (self.width, self.height))
        self.video = cv2.VideoWriter("my_video.avi", 0, 60, (self.width, self.height))

        # This is just used to count frames so we can cut it off at some point
        self.frames = 0

    def on_draw(self):
        self.clear()

    def on_update(self, delta_time):
        self.frames += 1
        img = self.get_image()
        self.video.write(img)
        if self.frames >= 1000:
            self.video.release()
            cv2.destroyAllWindows()
            arcade.exit()

        # make something happen for our video
        arcade.set_background_color(arcade.types.Color(self.frames % 255, 0, 0))

    # This is sort of a re-implementation of arcade.get_image(). Really the only difference
    # is in the last line, where we create a numpy array with the buffer instead of
    # creating a PIL image, and the fact that we are just capturing the whole screen instead
    # of allowing to capture a specific sub-section.
    def get_image(self):
        image_buffer = (gl.GLubyte * (3 * self.width * self.height))(0)
        gl.glReadPixels(0, 0, self.width, self.height, gl.GL_BGR, gl.GL_UNSIGNED_BYTE, image_buffer)
        return numpy.frombuffer(image_buffer, dtype="uint8").reshape(self.height, self.width, 3)  # type: ignore


def main():
    VideoRecorderCV2(SCREEN_WIDTH, SCREEN_HEIGHT, SCREEN_TITLE)
    arcade.run()


if __name__ == "__main__":
    main()



from abc import abstractmethod
from collections.abc import Set
from math import trunc
from typing import Protocol

from arcade.sprite import SpriteType, SpriteType_co
from arcade.sprite.base import BasicSprite
from arcade.types import IPoint, Point
from arcade.types.rect import Rect


class ReadOnlySpatialHash(Protocol[SpriteType_co]):
    """A read-only view of a :py:class:`.SpatialHash` which helps preserve safety.

    This works like the read-only views of Python's built-in :py:class:`dict`
    and other types. As an every-day user, it means that the underlying
    `SpatialHash` may contain subclasses of the annotated type, but not
    superclasses.

    This ensures predicable behavior via type safety in cases where:

    #. A spatial hash is annotated with a specific type
    #. It is then manipulated outside the original context with a broader type

    Advanced users who want more information on the specifics should see the
    comments of :py:class:`~arcade.sprite_list.SpriteList`.
    """

    @abstractmethod
    def get_sprites_near_sprite(self, sprite: BasicSprite) -> Set[SpriteType_co]:
        """
        Get all the sprites that are in the same buckets as the given sprite.

        Args:
            sprite: The sprite to check
        """
        ...

    @abstractmethod
    def get_sprites_near_point(self, point: Point) -> Set[SpriteType_co]:
        """
        Return sprites in the same bucket as the given point.

        Args:
            point: The point to check
        """
        ...

    @abstractmethod
    def get_sprites_near_rect(self, rect: Rect) -> Set[SpriteType_co]:
        """
        Return sprites in the same buckets as the given rectangle.

        .. tip:: Use :py:mod:`arcade.types.rect`'s helper functions to create
          rectangle objects!

        Args:
            rect:
                The rectangle to check as a :py:class:`~arcade.types.rect.Rect`
                object.
        """
        ...


class SpatialHash(ReadOnlySpatialHash[SpriteType]):
    """A data structure best for collision checks with non-moving sprites.

    It subdivides space into a grid of squares, each with sides of length
    :py:attr:`cell_size`. Moving a sprite from one place to another is the
    same as removing and adding it. Although moving a few can be okay, it
    can quickly add up and slow down a game.

    Args:
        cell_size:
            The width and height of each square in the grid.
    """

    def __init__(self, cell_size: int) -> None:
        # Sanity check the cell size
        if not isinstance(cell_size, int):
            raise TypeError("cell_size must be an int (integer)")
        if cell_size <= 0:
            raise ValueError("cell_size must be greater than 0")

        self.cell_size: int = cell_size
        """How big each grid cell is on each side.

        .. warning:: Do not change this after creation!

        Since each cell is a square, they're used as both the
        width and height.
        """
        # Buckets of sprites per cell
        self.contents: dict[IPoint, set[SpriteType]] = {}
        # All the buckets a sprite is in.
        # This is used to remove a sprite from the spatial hash.
        self.buckets_for_sprite: dict[SpriteType, list[set[SpriteType]]] = {}

    def hash(self, point: IPoint) -> IPoint:
        """Convert world coordinates to cell coordinates"""
        return (
            point[0] // self.cell_size,
            point[1] // self.cell_size,
        )

    def reset(self):
        """Clear all the sprites from the spatial hash."""
        self.contents.clear()
        self.buckets_for_sprite.clear()

    def add(self, sprite: SpriteType) -> None:
        """
        Add a sprite to the spatial hash.

        Args:
            sprite: The sprite to add
        """
        min_point = trunc(sprite.left), trunc(sprite.bottom)
        max_point = trunc(sprite.right), trunc(sprite.top)

        # hash the minimum and maximum points
        min_point, max_point = self.hash(min_point), self.hash(max_point)
        buckets: list[set[SpriteType]] = []

        # Iterate over the rectangular region adding the sprite to each cell
        for i in range(min_point[0], max_point[0] + 1):
            for j in range(min_point[1], max_point[1] + 1):
                # Add sprite to the bucket
                bucket = self.contents.setdefault((i, j), set())
                bucket.add(sprite)
                # Collect all the buckets we added to
                buckets.append(bucket)

        # Keep track of which buckets the sprite is in
        self.buckets_for_sprite[sprite] = buckets

    def move(self, sprite: SpriteType) -> None:
        """
        Shortcut to remove and re-add a sprite.

        Args:
            sprite: The sprite to move
        """
        self.remove(sprite)
        self.add(sprite)

    def remove(self, sprite: SpriteType) -> None:
        """
        Remove a Sprite.

        Args:
            sprite: The sprite to remove
        """
        # Remove the sprite from all the buckets it is in
        for bucket in self.buckets_for_sprite[sprite]:
            bucket.remove(sprite)

        # Delete the sprite from the bucket tracker
        del self.buckets_for_sprite[sprite]

    def get_sprites_near_sprite(self, sprite: BasicSprite) -> set[SpriteType]:
        min_point = trunc(sprite.left), trunc(sprite.bottom)
        max_point = trunc(sprite.right), trunc(sprite.top)

        # hash the minimum and maximum points
        min_point, max_point = self.hash(min_point), self.hash(max_point)
        close_by_sprites: set[SpriteType] = set()

        # Iterate over the all the covered cells and collect the sprites
        for i in range(min_point[0], max_point[0] + 1):
            for j in range(min_point[1], max_point[1] + 1):
                bucket = self.contents.setdefault((i, j), set())
                close_by_sprites.update(bucket)

        return close_by_sprites

    def get_sprites_near_point(self, point: Point) -> set[SpriteType]:
        hash_point = self.hash((trunc(point[0]), trunc(point[1])))
        # Return a copy of the set.
        return set(self.contents.setdefault(hash_point, set()))

    def get_sprites_near_rect(self, rect: Rect) -> set[SpriteType]:
        left, right, bottom, top = rect.lrbt
        min_point = trunc(left), trunc(bottom)
        max_point = trunc(right), trunc(top)

        # hash the minimum and maximum points
        min_point, max_point = self.hash(min_point), self.hash(max_point)
        close_by_sprites: set[SpriteType] = set()

        # Iterate over the all the covered cells and collect the sprites
        for i in range(min_point[0], max_point[0] + 1):
            for j in range(min_point[1], max_point[1] + 1):
                bucket = self.contents.setdefault((i, j), set())
                close_by_sprites.update(bucket)

        return close_by_sprites

    @property
    def count(self) -> int:
        """Return the number of sprites in the spatial hash"""
        # NOTE: We should really implement __len__ but this means
        # changing the truthiness of the class instance.
        # if spatial_hash will be False if it is empty.
        # For backwards compatibility, we'll keep it as a property.
        return len(self.buckets_for_sprite)



"""
When doing offscreen rendering we sometimes also want anti-aliasing with MSAA.

Framebuffers can have attachments for MSAA, but these are somewhat harder
to read from with a shader. Instead we simply copy the MSAA framebuffer to
the screen directly. This is allowed to easily read from the MSAA framebuffer.

If Python and Arcade are installed, this example can be run from the command line with:
python -m arcade.examples.gl.multisample_fbo
"""

import math
import arcade
from arcade.color import WHITE

SAMPLES = 8


class MultisampleFramebuffer(arcade.Window):

    def __init__(self, width, height):
        super().__init__(width, height, "Multisampled Framebuffer", samples=SAMPLES)
        # Create a MSAA texture and framebuffer
        self.texture = self.ctx.texture(self.get_framebuffer_size(), samples=SAMPLES)
        self.fbo = self.ctx.framebuffer(color_attachments=[self.texture])

    def on_draw(self):
        self.clear()

        # Draw to MSAA framebuffer
        self.fbo.use()
        self.fbo.clear()
        arcade.draw_line(0, 0, self.width, self.height, WHITE)
        arcade.draw_line(0, self.height, self.width, 0, WHITE)
        arcade.draw_circle_outline(
            self.width / 2,
            self.height / 2,
            100 + math.sin(self.time) * 50,
            WHITE,
        )
        arcade.draw_circle_outline(
            self.width / 2,
            self.height / 2,
            200 + math.sin(self.time) * 50,
            WHITE,
        )

        # Activate screen and copy the MSAA framebuffer to it
        # NOTE: It might be safer with a shader
        self.ctx.screen.use()
        self.ctx.copy_framebuffer(self.fbo, self.ctx.screen)


MultisampleFramebuffer(800, 600).run()



import warnings

from pyglet.input import Controller

import arcade
from arcade import ControllerManager


class _WindowControllerBridge:
    """Translates controller events to UIEvents and passes them to the UIManager.

    Controller are automatically connected and disconnected.

    Controller events are consumed by the UIControllerBridge,
    if the UIEvent is consumed by the UIManager.

    This implicates, that the UIControllerBridge should be the first listener in the chain and
    that other systems should be aware, when not to act on events (like when the UI is active).
    """

    def __init__(self, window: arcade.Window):
        self.window = window

        self.cm = ControllerManager()
        self.cm.push_handlers(self)

        # bind to existing controllers
        for controller in self.cm.get_controllers():
            self.on_connect(controller)

    def on_connect(self, controller: Controller):
        controller.push_handlers(self)

        try:
            controller.open()
        except Exception as e:
            warnings.warn(f"Failed to open controller {controller}: {e}")

        self.window.dispatch_event("on_connect", controller)

    def on_disconnect(self, controller: Controller):
        controller.remove_handlers(self)

        try:
            controller.close()
        except Exception as e:
            warnings.warn(f"Failed to close controller {controller}: {e}")

        self.window.dispatch_event("on_disconnect", controller)

    # Controller input event mapping
    def on_stick_motion(self, controller: Controller, name, value):
        return self.window.dispatch_event("on_stick_motion", controller, name, value)

    def on_trigger_motion(self, controller: Controller, name, value):
        return self.window.dispatch_event("on_trigger_motion", controller, name, value)

    def on_button_press(self, controller: Controller, button):
        return self.window.dispatch_event("on_button_press", controller, button)

    def on_button_release(self, controller: Controller, button):
        return self.window.dispatch_event("on_button_release", controller, button)

    def on_dpad_motion(self, controller: Controller, value):
        return self.window.dispatch_event("on_dpad_motion", controller, value)


class ControllerWindow(arcade.Window):
    """A window that automatically opens and listens to controller events
    and dispatches them via on_... hooks."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.cb = _WindowControllerBridge(self)

    def get_controllers(self) -> list[Controller]:
        """Return a list of connected controllers."""
        return self.cb.cm.get_controllers()

    # Controller event mapping
    def on_connect(self, controller: Controller):
        """Called when a controller is connected.
        The controller is already opened and ready to be used.
        """
        pass

    def on_disconnect(self, controller: Controller):
        """Called when a controller is disconnected."""
        pass

    def on_stick_motion(self, controller: Controller, name, value):
        """Called when a stick is moved."""
        pass

    def on_trigger_motion(self, controller: Controller, name, value):
        """Called when a trigger is moved."""
        pass

    def on_button_press(self, controller: Controller, button):
        """Called when a button is pressed."""
        pass

    def on_button_release(self, controller: Controller, button):
        """Called when a button is released."""
        pass

    def on_dpad_motion(self, controller: Controller, value):
        """Called when the dpad is moved."""
        pass


ControllerWindow.register_event_type("on_connect")
ControllerWindow.register_event_type("on_disconnect")
ControllerWindow.register_event_type("on_stick_motion")
ControllerWindow.register_event_type("on_trigger_motion")
ControllerWindow.register_event_type("on_button_press")
ControllerWindow.register_event_type("on_button_release")
ControllerWindow.register_event_type("on_dpad_motion")


class ControllerView(arcade.View):
    """A view which predefines the controller event mapping methods.

    Can be used with a ControllerWindow to handle controller events."""

    def on_connect(self, controller: Controller):
        """Called when a controller is connected.
        The controller is already opened and ready to be used.
        """
        pass

    def on_disconnect(self, controller: Controller):
        """Called when a controller is disconnected."""
        pass

    def on_stick_motion(self, controller: Controller, name, value):
        """Called when a stick is moved."""
        pass

    def on_trigger_motion(self, controller: Controller, name, value):
        """Called when a trigger is moved."""
        pass

    def on_button_press(self, controller: Controller, button):
        """Called when a button is pressed."""
        pass

    def on_button_release(self, controller: Controller, button):
        """Called when a button is released."""
        pass

    def on_dpad_motion(self, controller: Controller, value):
        """Called when the dpad is moved."""
        pass



"""
Using a Vertex Buffer Object With Lines

If Python and Arcade are installed, this example can be run from the command line with:
python -m arcade.examples.lines_buffered
"""
import random
import arcade
from arcade.shape_list import (
    ShapeElementList,
    create_line_strip,
)
from inspect import getmembers
from arcade.types import Color

# Do the math to figure out our screen dimensions
WINDOW_WIDTH = 1280
WINDOW_HEIGHT = 720
WINDOW_TITLE = "Vertex Buffer Object With Lines Example"


class GameView(arcade.View):
    """
    Main application class.
    """

    def __init__(self):
        """
        Set up the application.
        """
        super().__init__()

        self.shape_list = ShapeElementList()
        point_list = ((0, 50),
                      (10, 10),
                      (50, 0),
                      (10, -10),
                      (0, -50),
                      (-10, -10),
                      (-50, 0),
                      (-10, 10),
                      (0, 50))

        # Filter out anything other than a Color, such as imports and
        # helper functions.
        colors = [
            color for name, color in
            getmembers(arcade.color, lambda c: isinstance(c, Color))]

        for i in range(200):
            x = WINDOW_WIDTH // 2 - random.randrange(WINDOW_WIDTH)
            y = WINDOW_HEIGHT // 2 - random.randrange(WINDOW_HEIGHT)
            color = random.choice(colors)
            points = [(px + x, py + y) for px, py in point_list]

            my_line_strip = create_line_strip(points, color, 5)
            self.shape_list.append(my_line_strip)

        self.shape_list.center_x = WINDOW_WIDTH // 2
        self.shape_list.center_y = WINDOW_HEIGHT // 2
        self.shape_list.angle = 0

        self.background_color = arcade.color.BLACK

    def on_draw(self):
        """
        Render the screen.
        """
        # This command has to happen before we start drawing
        self.clear()

        self.shape_list.draw()

    def on_update(self, delta_time):
        self.shape_list.angle += 1 * 60 * delta_time
        self.shape_list.center_x += 0.1 * 60 * delta_time
        self.shape_list.center_y += 0.1 * 60 * delta_time


def main():
    """ Main function """
    # Create a window class. This is what actually shows up on screen
    window = arcade.Window(WINDOW_WIDTH, WINDOW_HEIGHT, WINDOW_TITLE, vsync=True)

    # Create and setup the GameView
    game = GameView()

    # Show GameView on screen
    window.show_view(game)

    # Start the arcade game loop
    arcade.run()


if __name__ == "__main__":
    main()



"""Sizing widgets using size hint keyword arguments

The following keyword arguments can be used to set preferred size
information for layouts which arrange widgets

* size_hint
* size_hint_max
* size_hint_min

Please note the following:

* These do nothing outside a layout
* They are only hints, and do not guarantee that a specific size will
  be provided.

If Arcade and Python are properly installed, you can run this example with:
python -m arcade.examples.gui.6_size_hints
"""

import textwrap

import arcade
from arcade.gui import (
    UIAnchorLayout,
    UIBoxLayout,
    UILabel,
    UIOnChangeEvent,
    UISpace,
    UITextArea,
    UIView,
)

arcade.resources.load_kenney_fonts()

SIZE_HINT_TEXT = textwrap.dedent(
    """
    UIWidgets provide three properties,
    which are used by layouts to determine the size of a widget.

These properties are:

* size_hint - percentage of the layout size
* size_hint_max - maximum size in pixels
* size_hint_min - minimum size in pixels

Theses properties can be None, or a tuple of two values. The first value is
the width, and the second value is the height.

If a value is None, the layout will use the widget's natural size for that dimension.

    """.strip()
)


class MyView(UIView):
    def __init__(self):
        super().__init__()
        self.background_color = arcade.uicolor.BLUE_BELIZE_HOLE

        root = self.ui.add(UIAnchorLayout())
        content = root.add(UIBoxLayout(size_hint=(1, 1)), anchor_x="left", anchor_y="top")

        # title and information
        header_box = content.add(
            UIBoxLayout(space_between=5, align="left", size_hint=(1, 0)),
        )
        header_box.with_border(color=arcade.uicolor.DARK_BLUE_MIDNIGHT_BLUE)
        header_box.with_padding(all=10)

        title = header_box.add(UILabel("Size Hint Example", font_size=24, bold=True))
        header_box.add(UISpace(color=arcade.uicolor.WHITE_CLOUDS, height=2, width=title.width))

        # create text area and set the minimal size to the content size
        text = header_box.add(
            UITextArea(
                text=SIZE_HINT_TEXT,
                width=800,  # give text enough space to not wrap
                font_size=14,
                size_hint=(1, 1),
                bold=True,
            )
        )
        text.with_padding(top=10)
        text.size_hint_min = (
            None,
            text.layout.content_height + 50,
        )  # set minimal height to content height + padding

        # add interactive demo
        content_anchor = content.add(UIAnchorLayout())
        content_anchor.with_border(color=arcade.uicolor.DARK_BLUE_MIDNIGHT_BLUE)
        content_anchor.with_padding(left=10, bottom=10)

        center_box = content_anchor.add(
            UIBoxLayout(size_hint=(0.8, 0.5), align="left", space_between=20)
        )

        width_slider_box = center_box.add(UIBoxLayout(vertical=False, size_hint=(1, 0)))
        width_slider_box.add(UILabel("Modify size_hint:", bold=True))
        width_slider = width_slider_box.add(
            arcade.gui.UISlider(
                min_value=0, max_value=1, value=0, size_hint=None, height=30, step=0.1
            )
        )
        width_value = width_slider_box.add(UILabel(bold=True))

        content_anchor.add(UISpace(height=50))

        # create a horizontal UIBoxLayout to show the effect of the sliders
        demo_box = center_box.add(UIBoxLayout(vertical=False, size_hint=(0.8, 1)))
        demo_box.with_background(color=arcade.uicolor.GRAY_ASBESTOS)

        # create a dummy widget to show the effect of the sliders
        dummy1 = demo_box.add(UILabel(bold=True))
        dummy1.with_background(color=arcade.uicolor.YELLOW_ORANGE)
        dummy2 = demo_box.add(UILabel(bold=True))
        dummy2.with_background(color=arcade.uicolor.GREEN_EMERALD)

        def update_size_hint_value(value: float):
            width_value.text = f"({value:.2f})"
            dummy1.size_hint = (value, 1)
            dummy1.text = f"size_hint = ({value:.2f}, 1)"

            dummy2.size_hint = (1 - value, 1)
            dummy2.text = f"size_hint = ({1 - value:.2f}, 1)"

        @width_slider.event("on_change")
        def on_change(event: UIOnChangeEvent):
            update_size_hint_value(event.new_value)

        initial_value = 1
        width_slider.value = initial_value
        update_size_hint_value(initial_value)

        content_anchor.add(UISpace(height=20))

        self.ui.execute_layout()
        self.ui.debug()


def main():
    window = arcade.Window(title="UIExample: Size Hints")
    window.show_view(MyView())
    window.run()


if __name__ == "__main__":
    main()



"""
Load a Tiled map file

Artwork from: https://kenney.nl
Tiled available from: https://www.mapeditor.org/

If Python and Arcade are installed, this example can be run from the command line with:
python -m arcade.examples.sprite_tiled_map
"""

import time

import arcade

TILE_SCALING = 0.5
PLAYER_SCALING = 1

WINDOW_WIDTH = 1280
WINDOW_HEIGHT = 720
WINDOW_TITLE = "Sprite Tiled Map Example"
SPRITE_PIXEL_SIZE = 128
GRID_PIXEL_SIZE = SPRITE_PIXEL_SIZE * TILE_SCALING
CAMERA_PAN_SPEED = 0.30

# Physics
MOVEMENT_SPEED = 5
JUMP_SPEED = 23
GRAVITY = 1.1


class GameView(arcade.View):
    """Main application class."""

    def __init__(self):
        """
        Initializer
        """
        super().__init__()

        # Tilemap Object
        self.tile_map = None

        # Sprite lists
        self.player_list = None
        self.wall_list = None
        self.coin_list = None

        # Set up the player
        self.score = 0
        self.player_sprite = None

        self.physics_engine = None
        self.end_of_map = 0
        self.game_over = False
        self.last_time = None
        self.frame_count = 0

        # Cameras
        self.camera = None
        self.camera_bounds = None
        self.gui_camera = None

        # Text
        self.fps_text = arcade.Text(
            "",
            x=10,
            y=40,
            color=arcade.color.BLACK,
            font_size=14
        )
        self.distance_text = arcade.Text(
            "0.0",
            x=10,
            y=20,
            color=arcade.color.BLACK,
            font_size=14,
        )

    def setup(self):
        """Set up the game and initialize the variables."""

        # Sprite lists
        self.player_list = arcade.SpriteList()

        # Set up the player
        self.player_sprite = arcade.Sprite(
            ":resources:images/animated_characters/female_person/femalePerson_idle.png",
            scale=PLAYER_SCALING,
        )

        # Starting position of the player
        self.player_sprite.center_x = 196
        self.player_sprite.center_y = 270
        self.player_list.append(self.player_sprite)

        map_name = ":resources:/tiled_maps/map.json"

        layer_options = {
            "Platforms": {"use_spatial_hash": True},
            "Coins": {"use_spatial_hash": True},
        }

        # Read in the tiled map
        self.tile_map = arcade.load_tilemap(
            map_name, layer_options=layer_options, scaling=TILE_SCALING
        )
        self.end_of_map = self.tile_map.width * GRID_PIXEL_SIZE

        # Set wall and coin SpriteLists
        self.wall_list = self.tile_map.sprite_lists["Platforms"]
        self.coin_list = self.tile_map.sprite_lists["Coins"]

        # --- Other stuff
        # Set the background color
        if self.tile_map.background_color:
            self.window.background_color = self.tile_map.background_color

        # Keep player from running through the wall_list layer
        walls = [self.wall_list, ]
        self.physics_engine = arcade.PhysicsEnginePlatformer(
            self.player_sprite, walls, gravity_constant=GRAVITY
        )

        self.camera = arcade.Camera2D()
        self.gui_camera = arcade.Camera2D()

        # Use the tilemap to limit the camera's position
        # we don't offset the max_y position to give a better experience.
        max_x = GRID_PIXEL_SIZE * self.tile_map.width
        max_y = GRID_PIXEL_SIZE * self.tile_map.height
        self.camera_bounds = arcade.LRBT(
            self.window.width / 2.0, max_x - self.window.width / 2.0,
            self.window.height / 2.0, max_y
        )

        # Center camera on user
        self.pan_camera_to_user()

        self.game_over = False

    def on_draw(self):
        """
        Render the screen.
        """

        # This command has to happen before we start drawing
        self.camera.use()
        self.clear()

        # Start counting frames
        self.frame_count += 1

        # Draw all the sprites.
        self.player_list.draw()
        self.wall_list.draw()
        self.coin_list.draw()

        # Activate GUI camera for FPS, distance and hit boxes
        # This will adjust text position based on viewport
        self.gui_camera.use()

        # Calculate FPS if conditions are met
        if self.last_time and self.frame_count % 60 == 0:
            fps = round(1.0 / (time.time() - self.last_time) * 60)
            self.fps_text.text = f"FPS: {fps:3d}"

        # Draw FPS text
        self.fps_text.draw()

        # Get time for every 60 frames
        if self.frame_count % 60 == 0:
            self.last_time = time.time()

        # Enable to draw hit boxes
        # self.wall_list.draw_hit_boxes()
        # self.wall_list_objects.draw_hit_boxes()

        # Get distance and draw text
        distance = self.player_sprite.right
        self.distance_text.text = f"Distance: {distance}"
        self.distance_text.draw()

        # Draw game over text if condition met
        if self.game_over:
            arcade.draw_text(
                "Game Over",
                200,
                200,
                arcade.color.BLACK,
                30,
            )

    def on_key_press(self, key, modifiers):
        """
        Called whenever a key is pressed.
        """
        if key == arcade.key.UP:
            if self.physics_engine.can_jump():
                self.player_sprite.change_y = JUMP_SPEED
        elif key == arcade.key.LEFT:
            self.player_sprite.change_x = -MOVEMENT_SPEED
        elif key == arcade.key.RIGHT:
            self.player_sprite.change_x = MOVEMENT_SPEED

    def on_key_release(self, key, modifiers):
        """
        Called when the user presses a mouse button.
        """
        if key == arcade.key.LEFT or key == arcade.key.RIGHT:
            self.player_sprite.change_x = 0

    def on_update(self, delta_time):
        """Movement and game logic"""

        if self.player_sprite.right >= self.end_of_map:
            self.game_over = True

        # Call update on all sprites
        if not self.game_over:
            self.physics_engine.update()

        coins_hit = arcade.check_for_collision_with_list(
            self.player_sprite, self.coin_list
        )
        for coin in coins_hit:
            coin.remove_from_sprite_lists()
            self.score += 1

        # Pan to the user
        self.pan_camera_to_user(CAMERA_PAN_SPEED)

    def pan_camera_to_user(self, panning_fraction: float = 1.0):
        """ Manage Scrolling """
        self.camera.position = arcade.math.smerp_2d(
            self.camera.position,
            self.player_sprite.position,
            self.window.delta_time,
            panning_fraction,
        )

        self.camera.position = arcade.camera.grips.constrain_xy(
            self.camera.view_data,
            self.camera_bounds
        )


def main():
    window = arcade.Window(WINDOW_WIDTH, WINDOW_HEIGHT, WINDOW_TITLE)
    game = GameView()
    game.setup()

    window.show_view(game)
    window.run()


if __name__ == "__main__":
    main()



"""
Parallax scrolling layers move slower the "farther" away they are.

Use the right and left arrow keys to move the car.

Arcade's ParallaxGroup allows you to implement this technique quickly
to create more satisfying backgrounds to your games. The example below
demonstrates how to fake an endless world by adjusting ParallaxGroup's
position & offset values. For limited worlds or backgrounds, limit the
repositioning to only occur within certain bounds, or delete it.

If Python and Arcade are installed, this example can be run from the command line with:
python -m arcade.examples.background_parallax
"""

import arcade
import arcade.future.background as background

# How much we'll scale up our pixel art
PIXEL_SCALE = 3

# The original & scaled heights of our background layer image data in pixels.
ORIGINAL_BG_LAYER_HEIGHT_PX = 240
SCALED_BG_LAYER_HEIGHT_PX = ORIGINAL_BG_LAYER_HEIGHT_PX * PIXEL_SCALE


WINDOW_TITLE = "Background Group Example"
WINDOW_WIDTH = 1280
WINDOW_HEIGHT = SCALED_BG_LAYER_HEIGHT_PX


PLAYER_SPEED = 300  # The player's speed in pixels / second
CAMERA_SPEED = 0.1


class GameView(arcade.View):
    def __init__(self):
        super().__init__()

        # Set the background color to match the sky in the background images
        self.background_color = (162, 84, 162, 255)

        self.camera = arcade.Camera2D()

        # Create a background group to hold all the landscape's layers
        self.backgrounds = background.ParallaxGroup()

        # Calculate the current size of each background fill layer in pixels
        bg_layer_size_px = (WINDOW_WIDTH, SCALED_BG_LAYER_HEIGHT_PX)

        # Import the image data for each background layer.
        # Unlike sprites, the scale argument doesn't resize the layer
        # itself. Instead, it changes the zoom level, while depth
        # controls how fast each layer scrolls. This means you have to
        # pass a correct size value when adding a layer. We calculated
        # this above.
        self.backgrounds.add_from_file(
            ":resources:/images/miami_synth_parallax/layers/back.png",
            size=bg_layer_size_px,
            depth=10.0,
            scale=PIXEL_SCALE
        )
        self.backgrounds.add_from_file(
            ":resources:/images/miami_synth_parallax/layers/buildings.png",
            size=bg_layer_size_px,
            depth=5.0,
            scale=PIXEL_SCALE
        )
        self.backgrounds.add_from_file(
            ":resources:/images/miami_synth_parallax/layers/palms.png",
            size=bg_layer_size_px,
            depth=3.0,
            scale=PIXEL_SCALE
        )
        self.backgrounds.add_from_file(
            ":resources:/images/miami_synth_parallax/layers/highway.png",
            size=bg_layer_size_px,
            depth=1.0,
            scale=PIXEL_SCALE
        )

        # Load the car texture and create a flipped version
        self.car_texture_right = arcade.texture.load_texture(
            ":resources:/images/miami_synth_parallax/car/car-idle.png")
        self.car_texture_left = self.car_texture_right.flip_left_right()

        # Create & position the player sprite in the center of the camera's view
        self.player_sprite = arcade.Sprite(
            self.car_texture_right,
            center_x=self.camera.viewport_width // 2, center_y=-200.0, scale=PIXEL_SCALE
        )

        self.player_sprite.bottom = 0

        # Track the player's x velocity
        self.x_velocity = 0

    def pan_camera_to_player(self):
        # Move the camera toward the center of the player's sprite
        target_x = self.player_sprite.center_x
        self.camera.position = arcade.math.lerp_2d(
            self.camera.position,
            (target_x, self.height//2),
            CAMERA_SPEED
        )

    def on_update(self, delta_time: float):
        # Move the player in our infinite world
        self.player_sprite.center_x += self.x_velocity * delta_time
        self.pan_camera_to_player()

    def on_draw(self):

        # Set up our drawing
        self.clear()
        with self.camera.activate():
            # Store a reference to the background layers as shorthand
            bg = self.backgrounds

            # Fake an endless world with scrolling terrain
            # Try experimenting with commenting out 1 or both of the 2 lines
            # below to get an intuitive understanding of what each does!
            bg.offset = self.camera.bottom_left  # Fake depth by moving layers
            bg.pos = self.camera.bottom_left  # Follow the car to fake infinity

            # Draw the background & the player's car
            with self.window.ctx.enabled(self.window.ctx.BLEND):
                bg.draw()
                arcade.draw_sprite(self.player_sprite, pixelated=True)

    def update_car_direction(self):
        if self.x_velocity < 0:
            self.player_sprite.texture = self.car_texture_left
        elif self.x_velocity > 0:
            self.player_sprite.texture = self.car_texture_right

    def on_key_press(self, symbol: int, modifiers: int):
        if symbol == arcade.key.LEFT:
            self.x_velocity -= PLAYER_SPEED
            self.update_car_direction()
        elif symbol == arcade.key.RIGHT:
            self.x_velocity += PLAYER_SPEED
            self.update_car_direction()

    def on_key_release(self, symbol: int, modifiers: int):
        if symbol == arcade.key.LEFT:
            self.x_velocity += PLAYER_SPEED
            self.update_car_direction()
        elif symbol == arcade.key.RIGHT:
            self.x_velocity -= PLAYER_SPEED
            self.update_car_direction()

    def on_resize(self, width: int, height: int):
        super().on_resize(width, height)
        self.camera.match_window()
        full_width_size = (width, SCALED_BG_LAYER_HEIGHT_PX)

        # We can iterate through a background group,
        # but in the case of a parallax group the iter returns
        # both the Backgrounds and the depths. (tuple[Background, float])
        for layer, depth in self.backgrounds:
            layer.size = full_width_size


def main():
    """ Main function """
    # Create a window class. This is what actually shows up on screen
    window = arcade.Window(WINDOW_WIDTH, WINDOW_HEIGHT, WINDOW_TITLE, resizable=True)

    # Create the GameView
    game = GameView()

    # Show GameView on screen
    window.show_view(game)

    # Start the arcade game loop
    arcade.run()


if __name__ == "__main__":
    main()



import warnings
from types import EllipsisType

from pyglet.event import EVENT_HANDLED, EVENT_UNHANDLED
from pyglet.math import Vec2

import arcade
from arcade import MOUSE_BUTTON_LEFT
from arcade.gui.events import (
    UIControllerButtonPressEvent,
    UIControllerButtonReleaseEvent,
    UIControllerDpadEvent,
    UIControllerEvent,
    UIEvent,
    UIKeyPressEvent,
    UIKeyReleaseEvent,
    UIMousePressEvent,
    UIMouseReleaseEvent,
)
from arcade.gui.property import ListProperty, Property, bind
from arcade.gui.surface import Surface
from arcade.gui.widgets import FocusMode, UIInteractiveWidget, UIWidget
from arcade.gui.widgets.layout import UIAnchorLayout
from arcade.gui.widgets.slider import UIBaseSlider


class UIFocusable(UIWidget):
    """
    A widget that provides additional information about focus neighbors.

    Attributes:

    neighbor_up: The widget above this widget.
    neighbor_right: The widget right of this widget.
    neighbor_down: The widget below this widget.
    neighbor_left: The widget left of this widget.
    """

    focus_mode = FocusMode.ALL

    neighbor_up: UIWidget | None = None
    neighbor_right: UIWidget | None = None
    neighbor_down: UIWidget | None = None
    neighbor_left: UIWidget | None = None


class UIFocusMixin(UIWidget):
    """A group of widgets that can be focused.

    UIFocusGroup maintains two lists of widgets:
    - The list of focusable widgets.
    - The list of widgets within (normal widget children).

    Use `detect_focusable_widgets()` to automatically detect focusable widgets
    or explicitly use `add_widget()`.

    The Group can be navigated with the keyboard (TAB/ SHIFT + TAB) or controller (DPad).

    - DPAD: Navigate between focusable widgets. (up, down, left, right)
    - TAB: Navigate between focusable widgets.
    - 'A' Button or SPACE: Interact with the focused widget.

    """

    _focused_widget = Property[UIWidget | None](None)
    _focusable_widgets = ListProperty[UIWidget]()
    _interacting: UIWidget | None = None

    _debug = Property(False)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        bind(self, "_debug", self.trigger_full_render)
        bind(self, "_focused_widget", self.trigger_full_render)
        bind(self, "_focusable_widgets", self.trigger_full_render)

    def on_event(self, event: UIEvent) -> bool | None:
        # pass events to children first, including controller events
        # so they can handle them
        if super().on_event(event):
            return EVENT_HANDLED

        if isinstance(event, UIControllerEvent):
            # if no focused widget, set the first focusable widget
            if self.focused_widget is None and self._focusable_widgets:
                self.set_focus()
                return EVENT_HANDLED

        if isinstance(event, UIKeyPressEvent):
            if event.symbol == arcade.key.TAB:
                if event.modifiers & arcade.key.MOD_SHIFT:
                    self.focus_previous()
                else:
                    self.focus_next()

                return EVENT_HANDLED

            elif event.symbol == arcade.key.SPACE:
                self._start_interaction()
                return EVENT_HANDLED

        elif isinstance(event, UIKeyReleaseEvent):
            if event.symbol == arcade.key.SPACE:
                self._end_interaction()
                return EVENT_HANDLED

        elif isinstance(event, UIControllerDpadEvent):
            if self._interacting:
                # TODO this should be handled in the slider!
                # pass dpad events to the interacting widget
                if event.vector.x == 1 and isinstance(self._interacting, UIBaseSlider):
                    self._interacting.norm_value += 0.1
                    return EVENT_HANDLED

                elif event.vector.x == -1 and isinstance(self._interacting, UIBaseSlider):
                    self._interacting.norm_value -= 0.1
                    return EVENT_HANDLED

                return EVENT_HANDLED

            else:
                # switch focus
                if event.vector.x == 1:
                    self.focus_right()
                    return EVENT_HANDLED

                elif event.vector.y == 1:
                    self.focus_up()
                    return EVENT_HANDLED

                elif event.vector.x == -1:
                    self.focus_left()
                    return EVENT_HANDLED

                elif event.vector.y == -1:
                    self.focus_down()
                    return EVENT_HANDLED

        elif isinstance(event, UIControllerButtonPressEvent):
            if event.button == "a":
                self._start_interaction()
                return EVENT_HANDLED
        elif isinstance(event, UIControllerButtonReleaseEvent):
            if event.button == "a":
                self._end_interaction()
                return EVENT_HANDLED

        return EVENT_UNHANDLED

    @classmethod
    def _walk_widgets(cls, root: UIWidget):
        for child in reversed(root.children):
            yield child
            yield from cls._walk_widgets(child)

    def detect_focusable_widgets(self):
        """Automatically detect focusable widgets."""
        widgets = self._walk_widgets(self)

        focusable_widgets = []
        for widget in reversed(list(widgets)):
            if self.is_focusable(widget):
                focusable_widgets.append(widget)

        self._focusable_widgets = focusable_widgets

    @property
    def focused_widget(self) -> UIWidget | None:
        """Return the currently focused widget.
        If no widget is focused, return None."""
        return self._focused_widget

    def set_focus(self, widget: UIWidget | None | EllipsisType = ...):
        """Set the focus to a specific widget.

        Set the focus to a specific widget. The widget must be in the list of
        focusable widgets. If the widget is not in the list, a ValueError is raised.

        Setting the focus to None will remove the focus from the current widget.
        If `...` is passed (default), the focus will be set to the first
        focusable widget in the list.

        Args:
            widget: The widget to focus.
        """
        # de-focus the current widget
        if widget is None:
            if self.focused_widget is not None:
                self.focused_widget.focused = False
                self._focused_widget = None
            return

        # resolve ...
        if widget is Ellipsis:
            if self._focusable_widgets:
                widget = self._focusable_widgets[0]
            else:
                raise ValueError(
                    "No focusable widgets in the group, "
                    "use `detect_focusable_widgets()` to detect them."
                )

        # handle new focus
        if widget not in self._focusable_widgets:
            raise ValueError("Widget is not focusable or not in the group.")

        if self.focused_widget is not None:
            self.focused_widget.focused = False
        widget.focused = True
        self._focused_widget = widget

    def focus_up(self):
        widget = self.focused_widget
        if isinstance(widget, UIFocusable):
            if widget.neighbor_up:
                self.set_focus(widget.neighbor_up)
                return

        self.focus_previous()

    def focus_down(self):
        widget = self.focused_widget
        if isinstance(widget, UIFocusable):
            if widget.neighbor_down:
                self.set_focus(widget.neighbor_down)
                return

        self.focus_next()

    def focus_left(self):
        widget = self.focused_widget
        if isinstance(widget, UIFocusable):
            if widget.neighbor_left:
                self.set_focus(widget.neighbor_left)
                return

        self.focus_previous()

    def focus_right(self):
        widget = self.focused_widget
        if isinstance(widget, UIFocusable):
            if widget.neighbor_right:
                self.set_focus(widget.neighbor_right)
                return

        self.focus_next()

    def focus_next(self):
        """Focus the next widget in the list of focusable widgets of this group"""
        if self.focused_widget is None:
            warnings.warn("No focused widget. Do not change focus.")
            return

        if self.focused_widget not in self._focusable_widgets:
            warnings.warn("Focused widget not in focusable widgets list. Do not change focus.")
            return

        focused_index = self._focusable_widgets.index(self.focused_widget) + 1
        focused_index %= len(self._focusable_widgets)  # wrap around
        self.set_focus(self._focusable_widgets[focused_index])

    def focus_previous(self):
        """Focus the previous widget in the list of focusable widgets of this group"""
        if self.focused_widget is None:
            warnings.warn("No focused widget. Do not change focus.")
            return

        if self.focused_widget not in self._focusable_widgets:
            warnings.warn("Focused widget not in focusable widgets list. Do not change focus.")
            return

        focused_index = self._focusable_widgets.index(self.focused_widget) - 1
        # automatically wrap around via index -1
        self.set_focus(self._focusable_widgets[focused_index])

    def _start_interaction(self):
        # TODO this should be handled in the widget

        widget = self.focused_widget

        if isinstance(widget, UIInteractiveWidget):
            widget.dispatch_ui_event(
                UIMousePressEvent(
                    source=self,
                    x=int(widget.rect.center_x),
                    y=int(widget.rect.center_y),
                    button=MOUSE_BUTTON_LEFT,
                    modifiers=0,
                )
            )
            self._interacting = widget
        else:
            print("Cannot interact widget")

    def _end_interaction(self):
        widget = self.focused_widget

        if isinstance(widget, UIInteractiveWidget):
            if isinstance(self._interacting, UIBaseSlider):
                # if slider, release outside the slider
                x = self._interacting.rect.left - 1
                y = self._interacting.rect.bottom - 1
            else:
                x = widget.rect.center_x
                y = widget.rect.center_y

            self._interacting = None
            widget.dispatch_ui_event(
                UIMouseReleaseEvent(
                    source=self,
                    x=int(x),
                    y=int(y),
                    button=MOUSE_BUTTON_LEFT,
                    modifiers=0,
                )
            )

    def _do_render(self, surface: Surface, force=False) -> bool:
        rendered = super()._do_render(surface, force)

        if rendered:
            self.do_post_render(surface)

        return rendered

    def do_post_render(self, surface: Surface):
        surface.limit(None)

        widget = self.focused_widget
        if not widget:
            return

        if self._debug:
            # debugging
            if isinstance(widget, UIFocusable):
                if widget.neighbor_up:
                    self._draw_indicator(
                        widget.rect.top_center,
                        widget.neighbor_up.rect.bottom_center,
                        color=arcade.color.RED,
                    )
                if widget.neighbor_down:
                    self._draw_indicator(
                        widget.rect.bottom_center,
                        widget.neighbor_down.rect.top_center,
                        color=arcade.color.GREEN,
                    )
                if widget.neighbor_left:
                    self._draw_indicator(
                        widget.rect.center_left,
                        widget.neighbor_left.rect.center_right,
                        color=arcade.color.BLUE,
                    )
                if widget.neighbor_right:
                    self._draw_indicator(
                        widget.rect.center_right,
                        widget.neighbor_right.rect.center_left,
                        color=arcade.color.ORANGE,
                    )

    def _draw_indicator(self, start: Vec2, end: Vec2, color=arcade.color.WHITE):
        arcade.draw_line(start.x, start.y, end.x, end.y, color, 2)
        arcade.draw_circle_filled(end.x, end.y, 5, color, num_segments=4)

    @staticmethod
    def is_focusable(widget):
        return widget.focus_mode is not FocusMode.NONE


class UIFocusGroup(UIFocusMixin, UIAnchorLayout):
    pass



"""
Array Backed Grid

Show how to use a two-dimensional list/array to back the display of a
grid on-screen. We can click each cell in the window to toggle the color
or each individual cell.

This is not the most efficient way to maintain an updated grid
simply because we have to rebuild the shape list from scratch
every time it changes, but it's fast enough for smaller grids
that don't update frequently.

If Python and Arcade are installed, this example can be run from the command line with:
python -m arcade.examples.array_backed_grid_buffered
"""
import arcade

# Set how many rows and columns we will have
ROW_COUNT = 15
COLUMN_COUNT = 15

# This sets the WIDTH and HEIGHT of each grid location
WIDTH = 30
HEIGHT = 30

# This sets the margin between each cell
# and on the edges of the screen.
MARGIN = 5

# Do the math to figure out our screen dimensions
WINDOW_WIDTH = (WIDTH + MARGIN) * COLUMN_COUNT + MARGIN
WINDOW_HEIGHT = (HEIGHT + MARGIN) * ROW_COUNT + MARGIN
WINDOW_TITLE = "Array Backed Grid Buffered Example"


class GameView(arcade.View):
    """
    Main application class.
    """

    def __init__(self):
        """
        Set up the application.
        """
        super().__init__()
        self.shape_list = None

        # Create a 2 dimensional array. A two dimensional
        # array is simply a list of lists.
        # This array can be altered later to contain 0 or 1
        # to show a white or green cell.
        #
        # A 4 x 4 grid would look like this
        #
        # grid = [
        #     [0, 0, 0, 0],
        #     [0, 0, 0, 0],
        #     [0, 0, 0, 0],
        #     [0, 0, 0, 0],
        # ]
        # We can quickly build a grid with python list comprehension
        # self.grid = [[0] * COLUMN_COUNT for _ in range(ROW_COUNT)]
        # Making the grid with loops:
        self.grid = []
        for row in range(ROW_COUNT):
            # Add an empty array that will hold each cell
            # in this row
            self.grid.append([])
            for column in range(COLUMN_COUNT):
                self.grid[row].append(0)  # Append a cell

        # Set the window's background color
        self.background_color = arcade.color.BLACK
        # Create shapes from the grid
        self.recreate_grid()

    def recreate_grid(self):
        """
        Create the shapes for our current grid.

        We look at the values in each cell.
        If the cell contains 0 we crate a white shape.
        If the cell contains 1 we crate a green shape.
        """
        self.shape_list = arcade.shape_list.ShapeElementList()
        for row in range(ROW_COUNT):
            for column in range(COLUMN_COUNT):
                if self.grid[row][column] == 0:
                    color = arcade.color.WHITE
                else:
                    color = arcade.color.GREEN

                x = (MARGIN + WIDTH) * column + MARGIN + WIDTH // 2
                y = (MARGIN + HEIGHT) * row + MARGIN + HEIGHT // 2

                current_rect = arcade.shape_list.create_rectangle_filled(
                    x, y, WIDTH, HEIGHT, color,
                )
                self.shape_list.append(current_rect)

    def on_draw(self):
        """
        Render the screen.
        """
        # We should always start by clearing the window pixels
        self.clear()

        # Draw the shapes representing our current grid
        self.shape_list.draw()

    def on_mouse_press(self, x, y, button, modifiers):
        """
        Called when the user presses a mouse button.
        """

        # Convert the clicked mouse position into grid coordinates
        column = int(x // (WIDTH + MARGIN))
        row = int(y // (HEIGHT + MARGIN))

        print(f"Click coordinates: ({x}, {y}). Grid coordinates: ({row}, {column})")

        # Make sure we are on-grid. It is possible to click in the upper right
        # corner in the margin and go to a grid location that doesn't exist
        if row >= ROW_COUNT or column >= COLUMN_COUNT:
            # Simply return from this method since nothing needs updating
            return

        # Flip the location between 1 and 0.
        if self.grid[row][column] == 0:
            self.grid[row][column] = 1
        else:
            self.grid[row][column] = 0

        # Rebuild the shapes
        self.recreate_grid()


def main():
    """ Main function """
    # Create a window class. This is what actually shows up on screen
    window = arcade.Window(WINDOW_WIDTH, WINDOW_HEIGHT, WINDOW_TITLE)

    # Create the GameView
    game = GameView()

    # Show GameView on screen
    window.show_view(game)

    # Start the arcade game loop
    arcade.run()


if __name__ == "__main__":
    main()



"""
Move Sprite by Angle

Simple program to show basic sprite usage.

Artwork from https://kenney.nl

If Python and Arcade are installed, this example can be run from the command line with:
python -m arcade.examples.sprite_move_angle
"""
import arcade
import math

SPRITE_SCALING = 0.5

WINDOW_WIDTH = 1280
WINDOW_HEIGHT = 720
WINDOW_TITLE = "Move Sprite by Angle Example"

MOVEMENT_SPEED = 5
ANGLE_SPEED = 5


class Player(arcade.Sprite):
    """ Player class """

    def __init__(self, image, scale):
        """ Set up the player """

        # Call the parent init
        super().__init__(image, scale=scale)

        # Create a variable to hold our speed. 'angle' is created by the parent
        self.speed = 0

    def update(self, delta_time: float = 1/60):
        # Rotate the ship
        self.angle += self.change_angle

        # Convert angle in degrees to radians.
        angle_rad = math.radians(self.angle)

        # Use math to find our change based on our speed and angle
        self.center_x += self.speed * math.sin(angle_rad)
        self.center_y += self.speed * math.cos(angle_rad)


class GameView(arcade.View):
    """
    Main application class.
    """

    def __init__(self):
        """
        Initializer
        """

        # Call the parent class initializer
        super().__init__()

        # Variables that will hold sprite lists
        self.player_list = None

        # Set up the player info
        self.player_sprite = None

        # Set the background color
        self.background_color = arcade.color.BLACK

    def setup(self):
        """ Set up the game and initialize the variables. """

        # Sprite lists
        self.player_list = arcade.SpriteList()

        # Set up the player
        self.player_sprite = Player(":resources:images/space_shooter/playerShip1_orange.png",
                                    SPRITE_SCALING)
        self.player_sprite.center_x = WINDOW_WIDTH / 2
        self.player_sprite.center_y = WINDOW_HEIGHT / 2
        self.player_list.append(self.player_sprite)

    def on_draw(self):
        """
        Render the screen.
        """

        # This command has to happen before we start drawing
        self.clear()

        # Draw all the sprites.
        self.player_list.draw()

    def on_update(self, delta_time):
        """ Movement and game logic """

        # Call update on all sprites (The sprites don't do much in this
        # example though.)
        self.player_list.update(delta_time)

    def on_key_press(self, key, modifiers):
        """Called whenever a key is pressed. """

        # Forward/back
        if key == arcade.key.UP:
            self.player_sprite.speed = MOVEMENT_SPEED
        elif key == arcade.key.DOWN:
            self.player_sprite.speed = -MOVEMENT_SPEED

        # Rotate left/right
        elif key == arcade.key.LEFT:
            self.player_sprite.change_angle = -ANGLE_SPEED
        elif key == arcade.key.RIGHT:
            self.player_sprite.change_angle = ANGLE_SPEED

    def on_key_release(self, key, modifiers):
        """Called when the user releases a key. """

        if key == arcade.key.UP or key == arcade.key.DOWN:
            self.player_sprite.speed = 0
        elif key == arcade.key.LEFT or key == arcade.key.RIGHT:
            self.player_sprite.change_angle = 0


def main():
    """ Main function """
    # Create a window class. This is what actually shows up on screen
    window = arcade.Window(WINDOW_WIDTH, WINDOW_HEIGHT, WINDOW_TITLE)

    # Create and setup the GameView
    game = GameView()
    game.setup()

    # Show GameView on screen
    window.show_view(game)

    # Start the arcade game loop
    arcade.run()


if __name__ == "__main__":
    main()



from pathlib import Path
from typing import TYPE_CHECKING

from arcade.texture.texture import Texture

if TYPE_CHECKING:
    from arcade import Texture
    from arcade.hitbox import HitBoxAlgorithm


class TextureBucket:
    """
    A simple dict based cache for textures.
    """

    def __init__(self):
        self._entries: dict[str, Texture] = {}

    def put(self, name: str, texture: Texture) -> None:
        """
        Add a texture to the cache.

        Args:
            name:
                The cache name of the texture
            texture:
                The texture to add
        """
        self._entries[name] = texture

    def get(self, name: str) -> Texture | None:
        """
        Get a texture from the cache by cache name.

        Args:
            name:
                The cache name of the texture
        Returns:
            The texture if found, otherwise ``None``
        """
        return self._entries.get(name)

    def delete(self, name: str, raise_if_not_exist: bool = True) -> None:
        """
        Delete a texture from the cache by cache name.

        Args:
            name:
                The cache name of the texture
            raise_if_not_exist:
                If ``True``, raises ``KeyError`` if the entry does not exist
        """
        try:
            del self._entries[name]
        except KeyError:
            if raise_if_not_exist:
                raise

    def delete_by_value(self, texture: "Texture") -> None:
        """
        Delete a texture from the cache by texture instance.

        Args:
            texture: The texture instance to delete
        """
        for name, value in self._entries.items():
            if value is texture:
                del self._entries[name]
                return

    def flush(self) -> None:
        """Clear the cache"""
        self._entries.clear()

    def __len__(self) -> int:
        return len(self._entries)


class TextureCache:
    """
    A cache for Arcade textures.

    The creation of a texture is an expensive operation for several reasons.
    * Loading an image from disk has a cost
    * Converting the image to a format suitable for OpenGL has a cost
    * Calculating the hash of the image has a cost
    * Creating the texture instance has a cost
    * Once the texture is created it's expensive to calculate hit box points

    This cache is intended to reduce the cost of creating textures by reusing
    existing ones. We also re-use the internal images on the existing textures
    for making different configurations of the same texture such as flipped
    and rotated versions including textures with different hit box configurations
    for the same image.
    """

    def __init__(self):
        self._entries = TextureBucket()
        self._file_entries = TextureBucket()

    def put(self, texture: "Texture") -> None:
        """
        Add a texture to the cache. It's important that the crop values
        and file path are correctly set on the texture before adding it to
        the cache.

        Args:
            texture: The texture to add
        """
        self._entries.put(texture.cache_name, texture)

        # Only cache by file path if it's the whole texture and not a crop
        # if texture.file_path and texture.crop_values in [None, (0, 0, 0, 0)]:
        #     self._file_entries.put(str(texture.file_path), texture)
        image_cache_name = texture.image_cache_name
        if image_cache_name:
            self._file_entries.put(image_cache_name, texture)

    def get(self, name: str) -> Texture | None:
        """
        Get a texture from the cache by cache name

        Args:
            name: The cache name of the texture
        Returns:
            The texture if found, otherwise ``None``
        """
        return self._entries.get(name)

    def get_with_config(self, hash: str, hit_box_algorithm: "HitBoxAlgorithm") -> Texture | None:
        """
        Attempts to find a texture with a specific configuration.

        Args:
            hash:
                The image hash
            hit_box_algorithm:
                The hit box algorithm to search for
        Returns:
            The texture if found, otherwise ``None``
        """
        from arcade import Texture

        name = Texture.create_cache_name(
            hash=hash,
            hit_box_algorithm=hit_box_algorithm,
        )
        return self.get(name)

    def get_texture_by_filepath(
        self,
        file_path: str | Path,
        crop: tuple[int, int, int, int] = (0, 0, 0, 0),
    ) -> Texture | None:
        """
        Get a texture from the cache by file path and crop values.

        Args:
            file_path: The path to the file the texture was loaded from
            crop: The crop values used when creating the texture
        """
        from arcade import Texture

        file_cache_name = Texture.create_image_cache_name(file_path, crop)
        return self._file_entries.get(file_cache_name)

    def delete(self, texture_or_name: Texture | str, raise_if_not_exist: bool = False) -> None:
        """
        Delete a texture from the cache by cache name.

        Args:
            texture_or_name:
                The texture or cache name to delete
            raise_if_not_exist:
                If ``True``, ignore errors when deleting
        """
        if isinstance(texture_or_name, Texture):
            texture = texture_or_name
            name = texture.cache_name
            # Delete from texture buckets
            self._entries.delete(name, raise_if_not_exist=raise_if_not_exist)
            # Delete from file bucket. Only present if file_path was provided
            self._file_entries.delete_by_value(texture)
        elif isinstance(texture_or_name, str):
            name = texture_or_name
            # Delete from texture buckets
            self._entries.delete(name, raise_if_not_exist=raise_if_not_exist)
            # Delete from file buckets
            self._file_entries.delete(name, raise_if_not_exist=False)
        else:
            raise TypeError(f"Expected Texture or str, got {type(texture_or_name)}")

    def flush(self) -> None:
        """Clear the cache"""
        self._entries.flush()
        self._file_entries.flush()

    def get_all_textures(self) -> set["Texture"]:
        """Get all textures in the cache"""
        return set(self._entries._entries.values())

    def __len__(self) -> int:
        """Count the number of unique textures"""
        return len(self._entries)

    def __contains__(self, texture: "Texture") -> bool:
        """Check if a texture is in the cache"""
        return texture in self.get_all_textures()

    def __iter__(self):
        """Iterate over all unique textures"""
        return iter(self.get_all_textures())

    def __getitem__(self, name: str) -> Texture | None:
        """Get a texture from the cache by cache name"""
        return self.get(name)

    def __setitem__(self, name: str, texture: "Texture") -> None:
        """Add a texture to the cache by cache name"""
        self.put(texture)

    def __delitem__(self, name: str) -> None:
        """Delete a texture from the cache by cache name"""
        self.delete(name, raise_if_not_exist=False)



"""
Builds on the custom sprite example adding bindless textures.

The old way of using textures is binding them to a texture unit
and assign that unit to a sampler in a shader. This is a global
state and can be a bit cumbersome to work with when you have
many textures and draw calls. It also means you have to build
texture atlases to properly batch draw calls. This comes with
additional complexity, limitations and challenges.

Bindless textures are textures referenced using a 64 bit integer
we call a "handle". This handle can be stored in a buffer and
declared as a sampler type (it's a 64 bit integer under the hood).
There is no need to bind the texture. They are all just available
though the handles.

This example builds on the custom sprite shader example.
We render points and use a geometry shader to expand them to
quads.

Note that texture handles are very unsafe. Using non-existing
handles or handles that are not resident can cause crashes.

If Python and Arcade are installed, this example can be run from the command line with:
python -m arcade.examples.gl.bindless_texture
"""

from array import array
from itertools import cycle

import arcade
from arcade.gl import BufferDescription, Texture2D


class BindlessTexture(arcade.Window):

    def __init__(self):
        super().__init__(
            1280,
            720,
            "Bindless Texture Example",
            resizable=True,
            gl_version=(4, 3),
        )

        # Custom sprite shader program using bindless textures
        self.program = self.ctx.program(
            vertex_shader="""
            #version 430

            // The point position
            in vec2 in_position;
            // Pass the vertex index (0, 1, ...) to the geometry shader
            flat out int v_texture_index;

            void main() {
                gl_Position = vec4(in_position, 0.0, 1.0);
                v_texture_index = gl_VertexID;
            }
            """,
            geometry_shader="""
            #version 430

            layout(points) in;
            layout(triangle_strip, max_vertices=4) out;

            // The global window block with matrices (always present)
            uniform WindowBlock {
                mat4 projection;
                mat4 view;
            } window;

            // Pick up the texture index from the vertex shader
            flat in int v_texture_index[];
            // Pass the texture index to the fragment shader
            flat out int texture_index;
            out vec2 uv;

            void main() {
                // Create some more convenient variables for the input
                vec2 center = gl_in[0].gl_Position.xy;
                vec2 hsize = vec2(40, 40);

                texture_index = v_texture_index[0];

                // Emit a triangle strip of 4 vertices making a triangle.
                // The fragment shader will then fill these to triangles in the next stage.
                mat4 mvp = window.projection * window.view;

                // Upper left
                gl_Position = mvp * vec4(vec2(-hsize.x, hsize.y) + center, 0.0, 1.0);
                uv = vec2(0, 1);
                EmitVertex();

                // lower left
                gl_Position =mvp * vec4(vec2(-hsize.x, -hsize.y) + center, 0.0, 1.0);
                uv = vec2(0, 0);
                EmitVertex();

                // upper right
                gl_Position = mvp * vec4(vec2(hsize.x, hsize.y) + center, 0.0, 1.0);
                uv = vec2(1, 1);
                EmitVertex();

                // lower right
                gl_Position = mvp * vec4(vec2(hsize.x, -hsize.y) + center, 0.0, 1.0);
                uv = vec2(1, 0);
                EmitVertex();

                // We are done with this triangle strip now
                EndPrimitive();
            }
            """,
            fragment_shader="""
            #version 460
            #extension GL_ARB_bindless_texture : require

            // This is the structure of the data in the shader storage buffer.
            // The texture handles in the buffer are 64 bit integers and are
            // automatically converted into sampler objects for us.
            struct TextureRef {
                layout (bindless_sampler) sampler2D tex; // 64 bit integer
            };

            // Shader storage buffer with texture handles
            layout(std430, binding=0) buffer BindlessData
            {
                TextureRef Materials[];
            } bindless_data;

            // Get the texture index from the geometry shader
            in flat int texture_index;
            in vec2 uv;
            out vec4 fragColor;

            void main() {
                // Look up the sampler in the shader storage buffer
                TextureRef ref = bindless_data.Materials[texture_index];
                // Write pixel to screen
                fragColor = texture(ref.tex, uv);
            }
            """,
        )

        self.handles = []
        self.textures: list[Texture2D] = []
        # Make a cycle iterator from Arcade's resources (images)
        resources = arcade.resources.list_built_in_assets(name="female", extensions=(".png",))
        resource_cycle = cycle(resources)

        # Load enough textures to cover for each point/sprite
        for i in range(16 * 9):
            texture = self.ctx.load_texture(
                next(resource_cycle),
                immutable=True,
                wrap_x=self.ctx.CLAMP_TO_EDGE,
                wrap_y=self.ctx.CLAMP_TO_EDGE,
            )
            # Make sure we keep a reference to the texture to avoid GC
            self.textures.append(texture)
            # Create a texture handle and make it resident.
            self.handles.append(texture.get_handle(resident=True))

        # Create the shader storage buffer with the texture handles
        self.texture_ssbo = self.ctx.buffer(data=array("Q", self.handles))

        # Generate some points/positions for our sprites
        pos = []
        for y in range(9):
            for x in range(16):
                pos.extend([x * 80 + 40, y * 80 + 40])

        # Create a buffer with the positions
        self.buffer_pos = self.ctx.buffer(data=array("f", pos))
        # Create a geometry object with the positions
        self.geometry = self.ctx.geometry(
            [BufferDescription(self.buffer_pos, "2f", ["in_position"])],
            mode=self.ctx.POINTS,
        )

    def on_draw(self):
        self.clear()
        self.ctx.enable(self.ctx.BLEND)
        # Bind the SSBO with texture handles to binding point 0
        # matching the binding point in the shader.
        self.texture_ssbo.bind_to_storage_buffer(binding=0)

        # Draw the sprites
        self.geometry.render(self.program)


BindlessTexture().run()



"""
Simple Pygame-Arcade interaction example

We render contents to a pygame surface and write that
to an OpenGL texture. Then we render that texture to
the screen using a simple texture shader. We use NEAREST
interpolation to keep the pixelated look.

This example was tested with Pygame Community Edition
but should work with any recent version of Pygame.

    pip install -I pygame-ce==2.1.3

"""

import math

import pygame  # type: ignore

import arcade
from arcade.gl import geometry


class SurfaceTexture:
    """
    Simple wrapper for a texture and a pygame surface
    making it simple to synchronize the data.
    """

    def __init__(self, ctx: arcade.ArcadeContext, size: tuple[int, int]):
        self._ctx = ctx
        self._size = size
        self._surface = pygame.Surface(size, flags=pygame.SRCALPHA)
        self._texture = self._ctx.texture(size, components=4)
        # Pygame surfaces are BGRA, so we need to swap the channels
        self._texture.swizzle = "BGRA"
        self._texture.filter = ctx.NEAREST, ctx.NEAREST
        self._geometry = geometry.quad_2d_fs()
        self._program = self._ctx.program(
            vertex_shader="""
            #version 330
            in vec2 in_vert;
            in vec2 in_uv;
            out vec2 uv;

            void main() {
                uv = in_uv;
                gl_Position = vec4(in_vert, 0.0, 1.0);
            }
            """,
            fragment_shader="""
            #version 330

            uniform sampler2D surface;
            in vec2 uv;
            out vec4 fragColor;

            void main() {
                // Flip texture y coordinate to get the right orientation
                fragColor = texture(surface, vec2(uv.x, 1 - uv.y));
            }
            """,
        )

    @property
    def size(self) -> tuple[int, int]:
        return self._size

    @property
    def surface(self) -> pygame.Surface:
        return self._surface

    def draw(self) -> None:
        """Draw the OpenGL texture to the screen"""
        self._texture.use(0)
        self._geometry.render(self._program)

    def write_surface_to_texture(self):
        """Update the texture with the surface data"""
        # This is the fastest way to write a surface to a texture.
        # Using pygame.image 100 x slower due to data conversion
        # and memory allocation.
        # The downside is that the raw surface data is RGBA, so we
        # set a swizzle on the OpenGL texture
        self._texture.write(self._surface.get_view("1"))


class PygameInteraction(arcade.Window):
    def __init__(self):
        super().__init__(512, 512, "Pygame Interaction")
        self.surface_texture = SurfaceTexture(self.ctx, size=(160, 160))

    def on_draw(self):
        self.clear()
        # Draw the pygame surface
        screen = self.surface_texture.surface
        w, h = self.surface_texture.size
        screen.fill((255, 255, 255))
        N = 8
        for i in range(N):
            time_offset = 6.28 / N * i
            pygame.draw.circle(
                screen,
                ((i * 50) % 255, (i * 100) % 255, (i * 20) % 255),
                (
                    math.sin(self.time + time_offset) * 55 + w // 2,
                    math.cos(self.time + time_offset) * 55 + h // 2,
                ),
                math.sin(self.time) * 4 + 15,
            )

        # Write the surface to the texture and draw it
        self.surface_texture.write_surface_to_texture()
        self.surface_texture.draw()

    def on_key_press(self, key, modifiers):
        if key == arcade.key.ESCAPE:
            arcade.close_window()


if __name__ == "__main__":
    PygameInteraction().run()



"""
Shows simple use of transform feedback.

Transforming is similar to rendering except that the output
or the shader is a buffer instead of a framebuffer/screen.

This examples shows a common ping-pong technique were we
transform a buffer with positions and velocities between
two buffers so we always work on the previous state.

* A list of N points are initialized with random positions and velocities
* A point of gravity is moving around on the screen affecting the points

Using transforms in this way makes us able to process
a system that is reacting to external forces in this way.
There are no predetermined paths and they system just lives on its own.

If Python and Arcade are installed, this example can be run from the command line with:
python -m arcade.examples.gl.transform_feedback
"""

from array import array
import math
import random
import arcade
import arcade.clock
from arcade.gl import BufferDescription

# Do the math to figure out our screen dimensions
WINDOW_WIDTH = 1280
WINDOW_HEIGHT = 720
WINDOW_TITLE = "Transform Feedback"


class GameView(arcade.View):

    def __init__(self):
        super().__init__()
        self.ctx = self.window.ctx

        # Program to visualize the points
        self.points_program = self.ctx.program(
            vertex_shader="""
            #version 330
            in vec2 in_pos;
            out vec3 color;
            void main() {
                // Let's just give them a "random" color based on the vertex id
                color = vec3(
                    mod(float(gl_VertexID * 100 % 11) / 10.0, 1.0),
                    mod(float(gl_VertexID * 100 % 27) / 10.0, 1.0),
                    mod(float(gl_VertexID * 100 % 71) / 10.0, 1.0));
                // Pass the point position to primitive assembly
                gl_Position = vec4(in_pos, 0.0, 1.0);
            }
            """,
            fragment_shader="""
            #version 330

            // Color passed in from the vertex shader
            in vec3 color;
            // The pixel we are writing to in the framebuffer
            out vec4 fragColor;

            void main() {
                // Fill the point
                fragColor = vec4(color, 1.0);
            }
            """,
        )

        # A program transforming points being affected by a gravity point
        self.gravity_program = self.ctx.program(
            vertex_shader="""
            #version 330

            // Delta time (since last frame)
            uniform float dt;
            // Strength of gravity
            uniform float force;
            // Position of gravity
            uniform vec2 gravity_pos;

            // The format of the data in our transform buffer(s)
            in vec2 in_pos;
            in vec2 in_vel;

            // We are writing to a buffer of the same format
            out vec2 out_pos;
            out vec2 out_vel;

            void main() {
                // Simplified gravity calculations
                vec2 dir = normalize(gravity_pos - in_pos) * force;
                vec2 vel = in_vel + dir / length(dir) * 0.01;

                // Write to the output buffer
                out_vel = vel;
                out_pos = in_pos + vel * dt;
            }
            """,
        )
        N = 50_000
        # Make two buffers we transform between so we can work on the previous result
        self.buffer_1 = self.ctx.buffer(data=array("f", self.gen_initial_data(N)))
        self.buffer_2 = self.ctx.buffer(reserve=self.buffer_1.size)

        # We also need to be able to visualize both versions (draw to the screen)
        self.vao_1 = self.ctx.geometry([BufferDescription(self.buffer_1, "2f 2x4", ["in_pos"])])
        self.vao_2 = self.ctx.geometry([BufferDescription(self.buffer_2, "2f 2x4", ["in_pos"])])

        # We need to be able to transform both buffers (ping-pong)
        self.gravity_1 = self.ctx.geometry(
            [BufferDescription(self.buffer_1, "2f 2f", ["in_pos", "in_vel"])]
        )
        self.gravity_2 = self.ctx.geometry(
            [BufferDescription(self.buffer_2, "2f 2f", ["in_pos", "in_vel"])]
        )

        self.ctx.enable_only()  # Ensure no context flags are set

    def gen_initial_data(self, count):
        for _ in range(count):
            yield random.uniform(-1.2, 1.2)  # pos x
            yield random.uniform(-1.2, 1.2)  # pos y
            yield random.uniform(-0.3, 0.3)  # velocity x
            yield random.uniform(-0.3, 0.3)  # velocity y

    def on_draw(self):
        self.clear()
        self.ctx.point_size = 2 * self.window.get_pixel_ratio()

        # Set uniforms in the program
        self.gravity_program["dt"] = self.window.delta_time
        self.gravity_program["force"] = 0.25
        self.gravity_program["gravity_pos"] = (
            math.sin(self.window.time * 0.77) * 0.25, math.cos(self.window.time) * 0.25
        )

        # Transform data in buffer_1 into buffer_2
        self.gravity_1.transform(self.gravity_program, self.buffer_2)
        # Render the result (Draw buffer_2)
        self.vao_2.render(self.points_program, mode=self.ctx.POINTS)

        # Swap around stuff around so we transform back and fourth between the two buffers
        self.gravity_1, self.gravity_2 = self.gravity_2, self.gravity_1
        self.vao_1, self.vao_2 = self.vao_2, self.vao_1
        self.buffer_1, self.buffer_2 = self.buffer_2, self.buffer_1


def main():
    """ Main function """
    # Create a window class. This is what actually shows up on screen
    window = arcade.Window(WINDOW_WIDTH, WINDOW_HEIGHT, WINDOW_TITLE)

    # Create the GameView
    game = GameView()

    # Show GameView on screen
    window.show_view(game)

    # Start the arcade game loop
    arcade.run()


if __name__ == "__main__":
    main()



from __future__ import annotations

import hashlib
from pathlib import Path
from typing import Any

import PIL.Image
import PIL.ImageDraw
import PIL.ImageOps

from arcade import hitbox
from arcade.color import TRANSPARENT_BLACK
from arcade.hitbox.base import HitBoxAlgorithm
from arcade.texture.transforms import (
    ORIENTATIONS,
    FlipLeftRightTransform,
    FlipTopBottomTransform,
    Rotate90Transform,
    Rotate180Transform,
    Rotate270Transform,
    Transform,
    TransposeTransform,
    TransverseTransform,
)
from arcade.types import RGBA255, Point2List

# from arcade.types.rect import Rect

__all__ = ["ImageData", "Texture"]


class ImageData:
    """
    A class holding the image for a texture with other metadata such as the hash.
    This information is used internally by the texture atlas to identify unique textures.

    If a hash is not provided, it will be calculated.
    By default, the hash is calculated using the sha256 algorithm.

    The ability to provide a hash directly is mainly there
    for ensuring we can load and save texture atlases to disk
    and for users to be able to allocate named regions in
    texture atlases.

    Args:
        image:
            The image for this texture
        hash:
            The hash of the image
    """

    __slots__ = ("image", "hash", "__weakref__")
    hash_func = "sha256"

    def __init__(self, image: PIL.Image.Image, hash: str | None = None, **kwargs):
        self.image = image
        """The pillow image"""
        self.hash = hash or self.calculate_hash(image)
        """The hash of the image"""

    @classmethod
    def calculate_hash(cls, image: PIL.Image.Image) -> str:
        """
        Calculates the hash of an image.

        The algorithm used is defined by the ``hash_func`` class variable.

        Args:
            image: The Pillow image to calculate the hash for
        """
        hash = hashlib.new(cls.hash_func)
        hash.update(image.tobytes())
        return hash.hexdigest()

    @property
    def width(self) -> int:
        """Width of the image in pixels."""
        return self.image.width

    @property
    def height(self) -> int:
        """Height of the image in pixels."""
        return self.image.height

    @property
    def size(self) -> tuple[int, int]:
        """The size of the image in pixels."""
        return self.image.size

    # ImageData uniqueness is based on the hash
    # -----------------------------------------
    def __hash__(self) -> int:
        return hash(self.hash)

    def __eq__(self, other) -> bool:
        if other is None:
            return False
        if not isinstance(other, self.__class__):
            return False
        return self.hash == other.hash

    def __ne__(self, other) -> bool:
        if other is None:
            return True
        if not isinstance(other, self.__class__):
            return True
        return self.hash != other.hash

    # -----------------------------------------

    def __repr__(self):
        return f"<ImageData width={self.width}, height={self.height}, hash={self.hash}>"


class Texture:
    """
    A Texture is wrapper for image data as a Pillow image,
    hit box data for this image used in collision detection
    and transformations like rotation and flipping if applied.

    Textures are by definition immutable. If you want to change
    the image data, you should create a new texture with the
    desired changes. There are some exceptions to this rule
    if you know the inner workings of the library.

    Args:
        image:
            The image or ImageData for this texture
        hit_box_algorithm:
            The algorithm to use for calculating the hit box.
        hit_box_points:
            A list of hitbox points for the texture to use (Optional).
            Completely overrides the hit box algorithm.
        hash:
            Optional unique name for the texture. Can be used to make this texture
            globally unique. By default the hash of the pixel data is used.
    """

    __slots__ = (
        "_image_data",
        "_size",
        "_vertex_order",
        "_transforms",
        "_hit_box_algorithm",
        "_hit_box_points",
        "_hash",
        "_cache_name",
        "_atlas_name",
        "_file_path",
        "_crop_values",
        "_properties",
        "__weakref__",
    )

    def __init__(
        self,
        image: PIL.Image.Image | ImageData,
        *,
        hit_box_algorithm: HitBoxAlgorithm | None = None,
        hit_box_points: Point2List | None = None,
        hash: str | None = None,
        **kwargs,
    ):
        # Overrides the hash
        self._hash = hash

        if isinstance(image, PIL.Image.Image):
            self._image_data = ImageData(image, hash=hash)
        elif isinstance(image, ImageData):
            self._image_data = image
        else:
            raise TypeError(
                f"image must be an instance of PIL.Image.Image or ImageData, not {type(image)}"
            )

        # Set the size of the texture since this is immutable
        self._size = image.width, image.height
        # The order of the texture coordinates when mapping
        # to a sprite/quad. This order is changed when the
        # texture is flipped or rotated.
        self._vertex_order = 0, 1, 2, 3

        self._hit_box_algorithm = hit_box_algorithm or hitbox.algo_default
        if not isinstance(self._hit_box_algorithm, HitBoxAlgorithm):
            raise TypeError(
                f"hit_box_algorithm must be an instance of HitBoxAlgorithm, "
                f"not {type(self._hit_box_algorithm)}"
            )

        # Internal names
        self._cache_name: str = ""
        self._atlas_name: str = ""
        self._update_cache_names()
        self._hit_box_points: Point2List = hit_box_points or self._calculate_hit_box_points()

        # Optional filename for debugging
        self._file_path: Path | None = None
        self._crop_values: tuple[int, int, int, int] | None = None
        self._properties: dict[str, Any] = {}

    @property
    def properties(self) -> dict[str, Any]:
        """
        A dictionary of properties for this texture.
        This can be used to store any data you want.
        """
        return self._properties

    @property
    def cache_name(self) -> str:
        """
        The name of the texture used for caching (read only).
        """
        return self._cache_name

    @property
    def image_cache_name(self) -> str | None:
        """
        Get the image cache name for this texture.
        Returns ``None`` if not loaded from a file.
        """
        if self.file_path is None:
            return None

        return self.create_image_cache_name(
            self.file_path,
            self.crop_values or (0, 0, 0, 0),
        )

    @classmethod
    def create_cache_name(
        cls,
        *,
        hash: str,
        hit_box_algorithm: HitBoxAlgorithm,
        vertex_order: tuple[int, int, int, int] = (0, 1, 2, 3),
    ) -> str:
        """
        Create a cache name for the texture.

        Args:
            hash:
                The hash of the image data
            hit_box_algorithm:
                The hit box algorithm for this texture
            vertex_order:
                The current vertex order of the texture
        """
        if not isinstance(hash, str):
            raise TypeError(f"Expected str, got {type(hash)}")
        if not isinstance(hit_box_algorithm, HitBoxAlgorithm):
            raise TypeError(f"Expected HitBoxAlgorithm, got {type(hit_box_algorithm)}")

        return f"{hash}|{vertex_order}|{hit_box_algorithm.cache_name}|"

    @classmethod
    def create_atlas_name(cls, hash: str, vertex_order: tuple[int, int, int, int] = (0, 1, 2, 3)):
        """
        Create a name for the texture in a texture atlas.

        Args:
            hash:
                The hash of the image data
            vertex_order:
                The current vertex order of the texture
        """
        return f"{hash}|{vertex_order}"

    def _update_cache_names(self):
        """Update the internal cache names."""
        self._cache_name = self.create_cache_name(
            hash=self._hash or self._image_data.hash,
            hit_box_algorithm=self._hit_box_algorithm,
            vertex_order=self._vertex_order,
        )
        self._atlas_name = self.create_atlas_name(
            hash=self._hash or self._image_data.hash,
            vertex_order=self._vertex_order,
        )

    @classmethod
    def create_image_cache_name(
        cls, path: str | Path, crop: tuple[int, int, int, int] = (0, 0, 0, 0)
    ):
        """
        Create a cache name for an image.

        Args:
            path:
                The path to the image file
            crop:
                The crop values used to create the texture
        """
        return f"{str(path)}|{crop}"

    @property
    def atlas_name(self) -> str:
        """
        The name of the texture used for the texture atlas (read only).
        """
        return self._atlas_name

    @property
    def file_path(self) -> Path | None:
        """
        A Path to the file this texture was loaded from
        """
        return self._file_path

    @file_path.setter
    def file_path(self, path: Path | None):
        self._file_path = path

    @property
    def crop_values(self) -> tuple[int, int, int, int] | None:
        """
        The crop values used to create this texture,
        This is used to track the real origin of the pixel data.
        """
        return self._crop_values

    @crop_values.setter
    def crop_values(self, crop: tuple[int, int, int, int] | None):
        self._crop_values = crop

    @property
    def image(self) -> PIL.Image.Image:
        """
        Get or set the image of the texture.

        .. warning::

            This is an advanced function. Be absolutely sure
            you know the consequences of changing the image.
            It can cause problems with the texture atlas and
            hit box points.
        """
        return self._image_data.image

    @image.setter
    def image(self, image: PIL.Image.Image):
        if image.size != self._image_data.image.size:
            raise ValueError("New image must be the same size as the old image")

        self._image_data.image = image

    @property
    def image_data(self) -> ImageData:
        """
        The image data of the texture (read only).

        This is a simple wrapper around the image containing metadata like hash
        and is used to determine the uniqueness of the image in texture atlases.
        """
        return self._image_data

    @property
    def width(self) -> int:
        """
        The virtual width of the texture in pixels.

        This can be different from the actual width of the image
        if set manually. It can be used to trick sprites into
        believing the texture us much larger than it is.
        """
        return self._size[0]

    @width.setter
    def width(self, value: int):
        self._size = (value, self._size[1])

    @property
    def height(self) -> int:
        """
        The virtual width of the texture in pixels.

        This can be different from the actual height of the image
        if set manually. It can be used to trick sprites into
        believing the texture us much larger than it is.
        """
        return self._size[1]

    @height.setter
    def height(self, value: int):
        self._size = (self._size[0], value)

    @property
    def size(self) -> tuple[int, int]:
        """
        The virtual size of the texture in pixels.

        This can be different from the actual size of the image
        if set manually. It can be used to trick sprites into
        believing the texture us much larger than it is.
        """
        return self._size

    @size.setter
    def size(self, value: tuple[int, int]):
        self._size = value

    @property
    def hit_box_points(self) -> Point2List:
        """
        Get the hit box points for this texture (read only).

        Custom hit box points must be supplied during texture creation
        and should ideally not be changed after creation.
        """
        return self._hit_box_points

    @property
    def hit_box_algorithm(self) -> HitBoxAlgorithm:
        """
        (read only) The algorithm used to calculate the hit box for this texture.
        """
        return self._hit_box_algorithm

    @classmethod
    def create_empty(
        cls,
        name: str,
        size: tuple[int, int],
        color: RGBA255 = TRANSPARENT_BLACK,
        hit_box_points: Point2List | None = None,
    ) -> Texture:
        """
        Create a texture with all pixels set to the given color.

        The hit box of the returned Texture will be set to a rectangle
        with the dimensions in ``size`` unless hit_box_points are provided.

        Args:
            name:
                The unique name for this texture. This is used for caching
                and uniqueness in texture atlases.
            size:
                The xy size of the internal image
            color:
                The color to fill the texture with
            hit_box_points:
                A list of hitbox points for the texture
        """
        return Texture(
            image=PIL.Image.new("RGBA", size, color),
            hash=name,
            hit_box_algorithm=hitbox.algo_bounding_box,
            hit_box_points=hit_box_points,
        )

    # ----- Transformations -----

    def flip_left_right(self) -> Texture:
        """
        Create a new texture that is flipped left to right from this texture.

        This returns a new texture with the same image data, but
        has updated hit box data and a transform that will be
        applied to the image when it's drawn (GPU side).
        """
        return self.transform(FlipLeftRightTransform)

    def flip_top_bottom(self) -> Texture:
        """
        Create a new texture that is flipped top to bottom from this texture.

        This returns a new texture with the same image data, but
        has updated hit box data and a transform that will be
        applied to the image when it's drawn (GPU side).
        """
        return self.transform(FlipTopBottomTransform)

    def flip_horizontally(self) -> Texture:
        """
        Create a new texture that is flipped horizontally from this texture.

        This returns a new texture with the same image data, but
        has updated hit box data and a transform that will be
        applied to the image when it's drawn (GPU side).
        """
        return self.flip_left_right()

    def flip_vertically(self) -> Texture:
        """
        Create a new texture that is flipped vertically from this texture.

        This returns a new texture with the same image data, but
        has updated hit box data and a transform that will be
        applied to the image when it's drawn (GPU side).
        """
        return self.flip_top_bottom()

    def flip_diagonally(self) -> Texture:
        """
        Creates a new texture that is flipped diagonally from this texture.
        This is an alias for :func:`transpose`.

        This returns a new texture with the same image data, but
        has updated hit box data and a transform that will be
        applied to the image when it's drawn (GPU side).
        """
        return self.transpose()

    def transpose(self) -> Texture:
        """
        Creates a new texture that is transposed from this texture.
        This flips the texture diagonally from lower right to upper left.

        This returns a new texture with the same image data, but
        has updated hit box data and a transform that will be
        applied to the image when it's drawn (GPU side).
        """
        return self.transform(TransposeTransform)

    def transverse(self) -> Texture:
        """
        Creates a new texture that is transverse from this texture.
        This flips the texture diagonally from lower left to upper right.

        This returns a new texture with the same image data, but
        has updated hit box data and a transform that will be
        applied to the image when it's drawn (GPU side).
        """
        return self.transform(TransverseTransform)

    def rotate_90(self, count: int = 1) -> Texture:
        """
        Create a new texture that is rotated 90 degrees from this texture.

        This returns a new texture with the same image data, but
        has updated hit box data and a transform that will be
        applied to the image when it's drawn (GPU side).

        Args:
            count: Number of 90 degree steps to rotate.
        """
        angles = [None, Rotate90Transform, Rotate180Transform, Rotate270Transform]
        count = count % 4
        transform = angles[count]
        if transform is None:
            return self
        return self.transform(transform)

    def rotate_180(self) -> Texture:
        """
        Create a new texture that is rotated 180 degrees from this texture.

        This returns a new texture with the same image data, but
        has updated hit box data and a transform that will be
        applied to the image when it's drawn (GPU side).
        """
        return self.transform(Rotate180Transform)

    def rotate_270(self) -> Texture:
        """
        Create a new texture that is rotated 270 degrees from this texture.

        This returns a new texture with the same image data, but
        has updated hit box data and a transform that will be
        applied to the image when it's drawn (GPU side).
        """
        return self.transform(Rotate270Transform)

    def transform(
        self,
        transform: type[Transform],
    ) -> Texture:
        """
        Create a new texture with the given transform applied.

        Args:
            transform: Transform to apply
        """
        new_hit_box_points = transform.transform_hit_box_points(self._hit_box_points)
        texture = Texture(
            self.image_data,
            hit_box_algorithm=self._hit_box_algorithm,
            hit_box_points=new_hit_box_points,
            hash=self._hash,
        )
        texture.width = self.width
        texture.height = self.height
        texture._vertex_order = transform.transform_vertex_order(self._vertex_order)
        texture.file_path = self.file_path
        texture.crop_values = self.crop_values

        # Swap width and height of the texture if needed
        old_rotation = ORIENTATIONS[self._vertex_order][0]
        new_rotation = ORIENTATIONS[texture._vertex_order][0]
        old_swapped = abs(old_rotation) % 180 != 0
        new_swapped = abs(new_rotation) % 180 != 0
        if old_swapped != new_swapped:
            texture.width, texture.height = self.height, self.width

        texture._update_cache_names()
        return texture

    def crop(
        self,
        x: int,
        y: int,
        width: int,
        height: int,
    ) -> Texture:
        """
        Create a new texture from a sub-section of this texture.

        If the crop is the same size as the original texture or
        the crop is 0 width or height, the original texture is
        returned.

        Args:
            x:
                X position to start crop
            y:
                Y position to start crop
            width:
                Width of crop
            height:
                Height of crop
        """
        # Return self if the crop is the same size as the original image
        if width == self.image.width and height == self.image.height and x == 0 and y == 0:
            return self

        # Return self width and height is 0
        if width == 0 and height == 0:
            return self

        self.validate_crop(self.image, x, y, width, height)

        area = (x, y, x + width, y + height)
        image = self.image.crop(area)
        image_data = ImageData(image)
        texture = Texture(
            image_data,
            hit_box_algorithm=self._hit_box_algorithm,
        )
        texture.crop_values = (x, y, width, height)
        return texture

    # ----- Utility functions -----

    @staticmethod
    def validate_crop(image: PIL.Image.Image, x: int, y: int, width: int, height: int) -> None:
        """
        Validate the crop values for a given image.

        Args:
            image:
                The image to crop
            x:
                X position to start crop
            y:
                Y position to start crop
            width:
                Width of crop
            height:
                Height of crop
        """
        if x < 0 or y < 0 or width < 0 or height < 0:
            raise ValueError(f"crop values must be positive: {x}, {y}, {width}, {height}")
        if x >= image.width:
            raise ValueError(f"x position is outside of texture: {x}")
        if y >= image.height:
            raise ValueError(f"y position is outside of texture: {y}")
        if x + width - 1 >= image.width:
            raise ValueError(f"width is outside of texture: {width + x}")
        if y + height - 1 >= image.height:
            raise ValueError(f"height is outside of texture: {height + y}")

    def _calculate_hit_box_points(self) -> Point2List:
        """
        Calculate the hit box points for this texture based on the configured
        hit box algorithm. This is usually done on texture creation
        or when the hit box points are requested the first time.
        """
        return self._hit_box_algorithm.calculate(self.image)



"""Points, sizes, and other similar aliases.

This is a submodule of :py:mod:`arcade.types` to avoid issues with:

* Circular imports
* Partially initialized modules

"""

from collections.abc import Sequence

from pyglet.math import Vec2, Vec3

from arcade.types.numbers import AsFloat

#: Matches both :py:class:`~pyglet.math.Vec2` and tuples of two numbers.
Point2 = tuple[AsFloat, AsFloat] | Vec2

#: Matches both :py:class:`~pyglet.math.Vec3` and tuples of three numbers.
Point3 = tuple[AsFloat, AsFloat, AsFloat] | Vec3


#: Matches any 2D or 3D point, including:
#:
#: * :py:class:`pyglet.math.Vec2`
#: * :py:class:`pyglet.math.Vec3`
#: * An ordinary :py:class:`tuple` of 2 or 3 values, either:
#:
#:   * :py:class:`int`
#    * :py:class:`float`
#:
#: This works the same way as :py:attr:`arcade.types.RGBOrA255` to
#: annotate RGB tuples, RGBA tuples, and :py:class:`tuple` or a
#: :py:class:`Color` instances.
Point = Point2 | Point3

PointList = Sequence[Point]
Point2List = Sequence[Point2]
Point3List = Sequence[Point3]


# Speed / typing workaround:
# 1. Eliminate extra allocations
# 2. Allows type annotation to be cleaner, primarily for HitBox & subclasses
EMPTY_POINT_LIST: Point2List = tuple()


class AnchorPoint:
    """Common anchor points as constants in UV space.

    Each is a :py:class:`~pyglet.math.Vec2` with axis values between
    ``0.0`` and ``1.0``. They can be used as arguments to
    :py:meth:`Rect.uv_to_position <arcade.types.Rect.uv_to_position>`
    to help calculate:

    * a pixel offset inside a :py:class:`~arcade.types.Rect`
    * an absolute screen positions in pixels

    Advanced users may also find them useful when working with
    shaders.
    """

    BOTTOM_LEFT = Vec2(0.0, 0.0)
    BOTTOM_CENTER = Vec2(0.5, 0.0)
    BOTTOM_RIGHT = Vec2(1.0, 0.0)
    CENTER_LEFT = Vec2(0.0, 0.5)
    CENTER = Vec2(0.5, 0.5)
    CENTER_RIGHT = Vec2(1.0, 0.5)
    TOP_LEFT = Vec2(0.0, 1.0)
    TOP_CENTER = Vec2(0.5, 1.0)
    TOP_RIGHT = Vec2(1.0, 1.0)


__all__ = [
    "Point2",
    "Point3",
    "Point",
    "Point2List",
    "Point3List",
    "PointList",
    "AnchorPoint",
    "EMPTY_POINT_LIST",
]



"""
A background group.

Background groups allow for many backgrounds to be drawn with a single draw call.
It also allows their texture scrolling together.
The position of the group is added to the position of each background.
This allows the backgrounds in the group to move together but not
necessarily positioned together.

If Python and Arcade are installed, this example can be run from the command line with:
python -m arcade.examples.background_groups
"""

import arcade
import arcade.future.background as background

WINDOW_WIDTH = 800
WINDOW_HEIGHT = 600

WINDOW_TITLE = "Background Group Example"

PLAYER_SPEED = 300
CAMERA_SPEED = 0.5


class GameView(arcade.View):
    def __init__(self):
        super().__init__()
        # Set the background color to equal to that of the first background.
        self.background_color = (5, 44, 70)

        self.camera = arcade.Camera2D()

        # create a background group which will hold all the backgrounds.
        self.backgrounds = background.BackgroundGroup()

        # Add each background from a file.
        # It is important to note that the scale only impacts the texture
        # and not the background. This means we need to ensure the background
        # size is also scaled correctly.
        self.backgrounds.add_from_file(
            ":resources:/images/cybercity_background/far-buildings.png",
            (0.0, 240.0),
            (WINDOW_WIDTH, 576),
            scale=3,
        )
        self.backgrounds.add_from_file(
            ":resources:/images/cybercity_background/back-buildings.png",
            (0.0, 120.0),
            (WINDOW_WIDTH, 576),
            scale=3,
        )
        self.backgrounds.add_from_file(
            ":resources:/images/cybercity_background/foreground.png",
            (0.0, 0.0),
            (WINDOW_WIDTH, 576),
            scale=3,
        )

        # Create the player sprite.
        self.player_sprite = arcade.SpriteSolidColor(20, 30, color=arcade.color.PURPLE)
        self.player_sprite.center_y = self.camera.viewport_height // 2
        self.player_sprite.center_x = self.camera.viewport_width // 2

        # Track Player Motion
        self.x_direction = 0
        self.y_direction = 0

    def pan_camera_to_player(self):
        # This will center the camera on the player.
        target_x = self.player_sprite.center_x
        target_y = self.player_sprite.center_y

        # This ensures the background is almost always at least partially visible.
        if 0.0 > target_x:
            target_x = 0.0
        elif target_x > 2.0 * self.camera.viewport_width:
            target_x = 2.0 * self.camera.viewport_width

        if 0.0 > target_y:
            target_y = 0.0
        elif target_y > 2.0 * self.camera.viewport_height:
            target_y = 2.0 * self.camera.viewport_height

        self.camera.position = arcade.math.lerp_2d(
            self.camera.position, (target_x, target_y), CAMERA_SPEED,
        )

    def on_update(self, delta_time: float):
        new_position = (
            self.player_sprite.center_x + self.x_direction * delta_time,
            self.player_sprite.center_y + self.y_direction * delta_time,
        )
        self.player_sprite.position = new_position

        self.pan_camera_to_player()

    def on_draw(self):
        self.clear()

        with self.camera.activate():
            with self.window.ctx.enabled(self.window.ctx.BLEND):
                self.backgrounds.draw()

            arcade.draw_sprite(self.player_sprite)

    def on_key_press(self, symbol: int, modifiers: int):
        # Support arrow keys and ASWD
        if symbol in (arcade.key.LEFT, arcade.key.A):
            self.x_direction -= PLAYER_SPEED
        elif symbol in (arcade.key.RIGHT, arcade.key.D):
            self.x_direction += PLAYER_SPEED
        elif symbol in (arcade.key.DOWN, arcade.key.S):
            self.y_direction -= PLAYER_SPEED
        elif symbol in (arcade.key.UP, arcade.key.W):
            self.y_direction += PLAYER_SPEED
        # Close the window if the user presses escape
        elif symbol == arcade.key.ESCAPE:
            self.window.close()

    def on_key_release(self, symbol: int, modifiers: int):
        if symbol in (arcade.key.LEFT, arcade.key.A):
            self.x_direction += PLAYER_SPEED
        elif symbol in (arcade.key.RIGHT, arcade.key.D):
            self.x_direction -= PLAYER_SPEED
        elif symbol in (arcade.key.DOWN, arcade.key.S):
            self.y_direction += PLAYER_SPEED
        elif symbol in (arcade.key.UP, arcade.key.W):
            self.y_direction -= PLAYER_SPEED

    def on_resize(self, width: int, height: int):
        super().on_resize(width, height)
        self.camera.match_window()


def main():
    """ Main function """
    # Create a window class. This is what actually shows up on screen
    window = arcade.Window(WINDOW_WIDTH, WINDOW_HEIGHT, WINDOW_TITLE, resizable=True)

    # Create and setup the GameView
    game = GameView()

    # Show GameView on screen
    window.show_view(game)

    # Start the arcade game loop
    arcade.run()


if __name__ == "__main__":
    main()



from arcade.shape_list import ShapeElementList, create_line
from arcade.types import RGBA255


def isometric_grid_to_screen(
    tile_x: int, tile_y: int, width: int, height: int, tile_width: int, tile_height: int
) -> tuple[int, int]:
    """
    Convert isometric grid coordinates to screen coordinates.

    Args:
        tile_x: The x coordinate of the tile in the isometric grid.
        tile_y: The y coordinate of the tile in the isometric grid.
        width: The width of the screen.
        height: The height of the screen.
        tile_width: The width of a tile in pixels.
        tile_height: The height of a tile in pixels.
    """
    screen_x = tile_width * tile_x // 2 + height * tile_width // 2 - tile_y * tile_width // 2
    screen_y = (
        (height - tile_y - 1) * tile_height // 2
        + width * tile_height // 2
        - tile_x * tile_height // 2
    )
    return screen_x, screen_y


def screen_to_isometric_grid(
    screen_x: int, screen_y: int, width: int, height: int, tile_width: int, tile_height: int
) -> tuple[int, int]:
    """
    Convert screen coordinates to isometric grid coordinates.

    Args:
        screen_x: The x coordinate on the screen.
        screen_y: The y coordinate on the screen.
        width: The width of the screen.
        height: The height of the screen.
        tile_width: The width of a tile in pixels.
        tile_height: The height of a tile in pixels.
    """
    x2 = (1 / tile_width * screen_x / 2 - 1 / tile_height * screen_y / 2 + width / 2) * 2 - (
        width / 2 + 0.5
    )
    y2 = (height - 1) - (
        (1 / tile_width * screen_x / 2 + 1 / tile_height * screen_y / 2) * 2 - (width / 2 + 0.5)
    )
    x2 = round(x2)
    y2 = round(y2)
    return x2, y2


def create_isometric_grid_lines(
    width: int, height: int, tile_width: int, tile_height: int, color: RGBA255, line_width: int
) -> ShapeElementList:
    """
    Create a ShapeElementList of isometric grid lines.

    Args:
        width: The width of the grid in tiles.
        height: The height of the grid in tiles.
        tile_width: The width of a tile in pixels.
        tile_height: The height of a tile in pixels.
        color: The color of the lines.
        line_width: The width of the lines.
    """
    # Grid lines 1
    shape_list: ShapeElementList = ShapeElementList()

    for tile_row in range(-1, height):
        tile_x = 0
        start_x, start_y = isometric_grid_to_screen(
            tile_x, tile_row, width, height, tile_width, tile_height
        )
        tile_x = width - 1
        end_x, end_y = isometric_grid_to_screen(
            tile_x, tile_row, width, height, tile_width, tile_height
        )

        start_x -= tile_width // 2
        end_y -= tile_height // 2

        line = create_line(start_x, start_y, end_x, end_y, color, line_width=line_width)
        shape_list.append(line)

    # Grid lines 2
    for tile_column in range(-1, width):
        tile_y = 0
        start_x, start_y = isometric_grid_to_screen(
            tile_column, tile_y, width, height, tile_width, tile_height
        )
        tile_y = height - 1
        end_x, end_y = isometric_grid_to_screen(
            tile_column, tile_y, width, height, tile_width, tile_height
        )

        start_x += tile_width // 2
        end_y -= tile_height // 2

        line = create_line(start_x, start_y, end_x, end_y, color, line_width=line_width)
        shape_list.append(line)

    return shape_list



"""Fundamental aliases, classes, and related constants.

As general rules:

#. Things only go in this module if they serve multiple purposes
   throughout arcade
#. Only expose the most important classes at this module's top level

For example, color-related types and related aliases go in
``arcade.types`` because they're used throughout the codebase. This
includes all the following areas:

#. :py:class:`~arcade.Sprite`
#. :py:class:`~arcade.SpriteList`
#. :py:class:`~arcade.Text`
#. The :py:mod:`arcade.gui` widgets
#. Functions in :py:mod:`arcade.drawing_commands`

However, since the color types, aliases, and constants are all related,
they go in the :py:mod:`arcade.types.color` submodule.
"""

# Don't lint import order since we have conditional compatibility shims
# flake8: noqa: E402
import sys
from pathlib import Path
from typing import Any, NamedTuple, TYPE_CHECKING, TypeVar, Protocol, Union
from collections.abc import Iterable

from pytiled_parser import Properties


# Backward-compatibility for buffer protocol objects
# To learn more, see https://docs.python.org/3.8/c-api/buffer.html
if sys.version_info >= (3, 12):
    from collections.abc import Buffer as BufferProtocol
else:
    # The most common built-in buffer protocol objects
    import ctypes
    from array import array
    from collections.abc import ByteString

    # This is used instead of the typing_extensions version since they
    # use an ABC which registers virtual subclasses. This will not work
    # with ctypes.Array since virtual subclasses must be concrete.
    # See: https://peps.python.org/pep-0688/
    BufferProtocol = ByteString | memoryview | array | ctypes.Array


# Since it's used everywhere, we'll prevent partial submodule
# and circular import problems by isolating it in a little box.
from arcade.types.numbers import AsFloat

# Generic color aliases
from arcade.types.color import RGB
from arcade.types.color import RGBA
from arcade.types.color import RGBOrA

# Specific color aliases
from arcade.types.color import RGB255
from arcade.types.color import RGBA255
from arcade.types.color import RGBNormalized
from arcade.types.color import RGBANormalized
from arcade.types.color import RGBOrA255
from arcade.types.color import RGBOrANormalized

# The Color helper type
from arcade.types.color import Color

# Vector-like items and collections
from arcade.types.vector_like import Point2
from arcade.types.vector_like import Point3
from arcade.types.vector_like import Point
from arcade.types.vector_like import Point2List
from arcade.types.vector_like import Point3List
from arcade.types.vector_like import PointList
from arcade.types.vector_like import EMPTY_POINT_LIST
from arcade.types.vector_like import AnchorPoint

# Rectangles
from arcade.types.rect import IntRectParams
from arcade.types.rect import RectParams
from arcade.types.rect import RectKwargs

from arcade.types.rect import Rect
from arcade.types.rect import LRBT
from arcade.types.rect import LBWH
from arcade.types.rect import XYWH
from arcade.types.rect import XYRR
from arcade.types.rect import Viewport

# Boxes
from arcade.types.box import Box
from arcade.types.box import LBNWHD
from arcade.types.box import LRBTNF
from arcade.types.box import XYZWHD


__all__ = [
    "AsFloat",
    "BufferProtocol",
    "Box",
    "XYZWHD",
    "LRBTNF",
    "LBNWHD",
    "Color",
    "IntRectParams",
    "IPoint",
    "PathOr",
    "PathOrTexture",
    "Point",
    "Point2",
    "Point3",
    "PointList",
    "Point2List",
    "Point3List",
    "OneOrIterableOf",
    "EMPTY_POINT_LIST",
    "AnchorPoint",
    "Rect",
    "LRBT",
    "LBWH",
    "XYWH",
    "XYRR",
    "Viewport",
    "RectParams",
    "RectKwargs",
    "HasAddSubMul",
    "RGB",
    "RGBA",
    "RGBOrA",
    "RGB255",
    "RGBA255",
    "RGBOrA255",
    "RGBNormalized",
    "RGBANormalized",
    "RGBOrANormalized",
    "Size2D",
    "TiledObject",
    "Velocity",
]


_T = TypeVar("_T")


OneOrIterableOf = Union[_T, Iterable[_T]]  # noqa: UP007
"""Either an instance of something or an iterable of them.

When writing loading code which is not performance critical,
you may also want to see the following:

* :py:func:`arcade.utils.grow_sequence`
* :py:func:`arcade.utils.is_iterable`
* :py:func:`arcade.utils.is_nonstr_iterable`
* :py:func:`arcade.utils.is_str_or_noniterable`

.. tip:: You can inline the contents of these functions when
         performance matters.

You should avoid using this annotation with values which may include
``None``. These include:

* Pipe syntax such as ``OneOrIterable[MyType | None]``
* :py:class:`typing.Optional` equivalents of pipe syntax

Instead, consider one of the following when possible:

.. code-block:: python

   def annotated(argument: OneOrIterableOf[MyType] = tuple()):
      ...

   def annotated2(argument: OneOrIterableOf[MyType] | None = tuple()):
      ...

"""

# --- Begin potentially obsolete annotations ---

#: ``Size2D`` helps mark int or float sizes. Use it like a
#: :py:class:`typing.Generic`'s bracket notation as follows:
#:
#: .. code-block:: python
#:
#:    def example_Function(size: Size2D[int], color: RGBA255) -> Texture:
#:       """An example of how to use Size2D.
#:
#:       Look at signature above, not the missing function body. The
#:       ``size`` argument is how you can mark texture sizes, while
#:       you can use ``Size2D[float]`` to denote float regions.
#:
#        Args:
#:           size: A made-up hypothetical argument.
#:           color: Hypothetical texture-related argument.
#:       """
#:       ...  # No function definition
#:
Size2D = tuple[_T, _T]

#: Used in :py:class:`~arcade.sprite_list.spatial_hash.SpatialHash`.
IPoint = tuple[int, int]


# We won't keep this forever. It's a temp stub for particles we'll replace.
Velocity = tuple[AsFloat, AsFloat]

# --- End potentially obsolete annotations ---


# These are for the argument type + return type. They're separate TypeVars
# to handle cases which take tuple but return Vec2 (e.g. pyglet.math.Vec2).
_T_contra = TypeVar("_T_contra", contravariant=True)  # Same or more general than T
_T_co = TypeVar("_T_co", covariant=True)  # Same or more specific than T


class HasAddSubMul(Protocol[_T_contra, _T_co]):
    """Matches types which work with :py:func:`arcade.math.lerp`."""

    # The / matches float and similar operations to keep pyright
    # happy since built-in arithmetic makes them positional only.
    # See https://peps.python.org/pep-0570/
    def __add__(self, value: _T_contra, /) -> _T_co: ...
    def __sub__(self, value: _T_contra, /) -> _T_co: ...
    def __mul__(self, value: _T_contra, /) -> _T_co: ...


# Path handling
PathLike = str | Path | bytes
_POr = TypeVar("_POr")  # Allows PathOr[TypeNameHere] syntax
PathOr = Union[PathLike, _POr]  # noqa: UP007


# Specific utility resource aliases with type imports
if TYPE_CHECKING:
    # The linters are wrong: this is used, so we noqa it
    from arcade.texture import Texture  # noqa: F401

PathOrTexture = PathOr["Texture"]


class TiledObject(NamedTuple):
    """Object in a tilemaps"""

    shape: Point | PointList | tuple[int, int, int, int]
    """Shape of the object"""
    properties: Properties | None = None
    """Properties of the object"""
    name: str | None = None
    """Name of the object"""
    type: str | None = None
    """Type of the object"""


# Stolen from Pylance
class SupportsDunderLT(Protocol[_T_contra]):
    def __lt__(self, other: _T_contra, /) -> bool: ...


class SupportsDunderGT(Protocol[_T_contra]):
    def __gt__(self, other: _T_contra, /) -> bool: ...


SupportsRichComparison = SupportsDunderLT[Any] | SupportsDunderGT[Any]



"""Customizing buttons with text & textures.

This example showcases Arcade's range of different built-in button types
and how they can be used to customize a UI. A UIGridLayout is used to
arrange buttons.

If Arcade and Python are properly installed, you can run this example with:
python -m arcade.examples.gui.3_buttons
"""

import arcade
from arcade.gui import (
    UIAnchorLayout,
    UIFlatButton,
    UIGridLayout,
    UIImage,
    UIOnChangeEvent,
    UITextureButton,
    UITextureToggle,
    UIView,
)


# Preload textures, because they are mostly used multiple times, so they are not
# loaded multiple times
ICON_SMALLER = arcade.load_texture(":resources:gui_basic_assets/icons/smaller.png")
ICON_LARGER = arcade.load_texture(":resources:gui_basic_assets/icons/larger.png")

TEX_SWITCH_GREEN = arcade.load_texture(":resources:gui_basic_assets/toggle/green.png")
TEX_SWITCH_RED = arcade.load_texture(":resources:gui_basic_assets/toggle/red.png")
TEX_RED_BUTTON_NORMAL = arcade.load_texture(":resources:gui_basic_assets/button/red_normal.png")
TEX_RED_BUTTON_HOVER = arcade.load_texture(":resources:gui_basic_assets/button/red_hover.png")
TEX_RED_BUTTON_PRESS = arcade.load_texture(":resources:gui_basic_assets/button/red_press.png")


class MyView(UIView):
    def __init__(self):
        super().__init__()
        self.background_color = arcade.uicolor.BLUE_PETER_RIVER

        grid = UIGridLayout(
            column_count=3,
            row_count=4,
            size_hint=(0, 0),
            vertical_spacing=10,
            horizontal_spacing=10,
        )

        self.ui.add(UIAnchorLayout(children=[grid]))

        # simple UIFlatButton with text
        grid.add(UIFlatButton(text="UIFlatButton", width=200), row=0, column=0)

        # UIFlatButton change text placement right
        flat_with_more_text = UIFlatButton(text="text placed right", width=200)
        flat_with_more_text.place_text(anchor_x="right")
        grid.add(flat_with_more_text, row=1, column=0)

        # UIFlatButton change text placement right
        flat_with_more_text = UIFlatButton(text="text placed top left", width=200)
        flat_with_more_text.place_text(anchor_x="left", anchor_y="top")
        grid.add(flat_with_more_text, row=2, column=0)

        # UIFlatButton with icon on the left
        flat_with_icon_left = UIFlatButton(text="UIFlatButton with icon", width=200)
        flat_with_icon_left.place_text(align_x=+20)
        flat_with_icon_left.add(
            child=UIImage(
                texture=ICON_LARGER,
                width=30,
                height=30,
            ),
            anchor_x="left",
            align_x=10,
        )
        grid.add(flat_with_icon_left, row=0, column=1)

        # UIFlatButton with icon on the right
        flat_with_icon_right = UIFlatButton(text="UIFlatButton with icon", width=200)
        flat_with_icon_right.place_text(align_x=-20)
        flat_with_icon_right.add(
            child=UIImage(
                texture=ICON_SMALLER,
                width=30,
                height=30,
            ),
            anchor_x="right",
            align_x=-10,
        )
        grid.add(flat_with_icon_right, row=1, column=1)

        # UIFlatButton with icon on both sides
        flat_with_icon_right = UIFlatButton(text="UIFlatButton", width=200)
        flat_with_icon_right.add(
            child=UIImage(
                texture=ICON_SMALLER,
                width=30,
                height=30,
            ),
            anchor_x="left",
            align_x=10,
        )
        flat_with_icon_right.add(
            child=UIImage(
                texture=ICON_SMALLER,
                width=30,
                height=30,
            ),
            anchor_x="right",
            align_x=-10,
        )
        grid.add(flat_with_icon_right, row=2, column=1)

        # UITextureButton
        texture_button = UITextureButton(
            text="UITextureButton",
            width=200,
            texture=TEX_RED_BUTTON_NORMAL,
            texture_hovered=TEX_RED_BUTTON_HOVER,
            texture_pressed=TEX_RED_BUTTON_PRESS,
        )
        grid.add(texture_button, row=0, column=2)

        # UITextureButton with icon left
        texture_button_with_icon_left = UITextureButton(
            text="UITextureButton",
            width=200,
            texture=TEX_RED_BUTTON_NORMAL,
            texture_hovered=TEX_RED_BUTTON_HOVER,
            texture_pressed=TEX_RED_BUTTON_PRESS,
        )
        texture_button_with_icon_left.add(
            child=UIImage(
                texture=ICON_SMALLER,
                width=25,
                height=25,
            ),
            anchor_x="left",
            align_x=10,
        )
        grid.add(texture_button_with_icon_left, row=1, column=2)

        # UITextureButton with multiline text
        texture_button_with_icon_left = UITextureButton(
            text="UITextureButton\nwith a second line",
            multiline=True,
            width=200,
            texture=TEX_RED_BUTTON_NORMAL,
            texture_hovered=TEX_RED_BUTTON_HOVER,
            texture_pressed=TEX_RED_BUTTON_PRESS,
        )
        texture_button_with_icon_left.place_text(anchor_x="left", align_x=45)
        texture_button_with_icon_left.add(
            child=UIImage(
                texture=ICON_SMALLER,
                width=25,
                height=25,
            ),
            anchor_x="left",
            align_x=10,
        )
        grid.add(texture_button_with_icon_left, row=2, column=2)

        # UIFlatButtons with toggle
        texture_button_with_toggle = UIFlatButton(
            text="Just get crazy now!",
            width=630,
        )
        texture_button_with_toggle.place_text(anchor_x="left", align_x=45)
        texture_button_with_toggle.add(
            child=UIImage(
                texture=ICON_SMALLER,
                width=25,
                height=25,
            ),
            anchor_x="left",
            align_x=10,
        )
        toggle = texture_button_with_toggle.add(
            child=UITextureToggle(
                on_texture=TEX_SWITCH_RED,
                off_texture=TEX_SWITCH_GREEN,
                width=60,
                height=30,
            ),
            anchor_x="right",
            align_x=-10,
        )

        @toggle.event
        def on_change(event: UIOnChangeEvent):
            texture_button_with_toggle.disabled = event.new_value

        grid.add(texture_button_with_toggle, row=3, column=0, column_span=3)


def main():
    window = arcade.Window(800, 600, "GUI Example: Buttons", resizable=True)
    window.show_view(MyView())
    window.run()


if __name__ == "__main__":
    main()



"""
Module for handling common texture transforms
such as rotation, translation, flipping etc.

We don't actually transform pixel data, we simply
transform the texture coordinates and hit box points.
"""

from enum import Enum

from arcade.math import rotate_point
from arcade.types import Point2List


class VertexOrder(Enum):
    """Order for texture coordinates."""

    UPPER_LEFT = 0
    UPPER_RIGHT = 1
    LOWER_LEFT = 2
    LOWER_RIGHT = 3


class Transform:
    """
    Base class for all texture transforms.

    Transforms are responsible for transforming the texture
    coordinates and hit box points.
    """

    #: How texture coordinates order should be changed
    #: for this transform.
    #: upper_left, upper_right, lower_left, lower_right
    order = (
        VertexOrder.UPPER_LEFT.value,
        VertexOrder.UPPER_RIGHT.value,
        VertexOrder.LOWER_LEFT.value,
        VertexOrder.LOWER_RIGHT.value,
    )

    @staticmethod
    def transform_hit_box_points(
        points: Point2List,
    ) -> Point2List:
        """Transforms hit box points."""
        return points

    @classmethod
    def transform_vertex_order(cls, order: tuple[int, int, int, int]) -> tuple[int, int, int, int]:
        """
        Transforms the vertex order with the incoming order.

        This gives us important metadata on how to quickly transform
        the texture coordinates without iterating all applied transforms.

        Args:
            order: The incoming order
        Returns:
            The new vertex order
        """
        return (
            order[cls.order[0]],
            order[cls.order[1]],
            order[cls.order[2]],
            order[cls.order[3]],
        )

    @classmethod
    def transform_texture_coordinates_order(
        cls,
        texture_coordinates: tuple[float, float, float, float, float, float, float, float],
        order: tuple[int, int, int, int],
    ) -> tuple[float, float, float, float, float, float, float, float]:
        """
        Change texture coordinates order by a vertex order.

        Args:
            texture_coordinates:
                Texture coordinates to transform
            order:
                The new order
        """
        uvs = texture_coordinates
        return (
            uvs[order[0] * 2],
            uvs[order[0] * 2 + 1],
            uvs[order[1] * 2],
            uvs[order[1] * 2 + 1],
            uvs[order[2] * 2],
            uvs[order[2] * 2 + 1],
            uvs[order[3] * 2],
            uvs[order[3] * 2 + 1],
        )


class Rotate90Transform(Transform):
    """
    Rotate 90 degrees clockwise.
    """

    order = (
        VertexOrder.LOWER_LEFT.value,
        VertexOrder.UPPER_LEFT.value,
        VertexOrder.LOWER_RIGHT.value,
        VertexOrder.UPPER_RIGHT.value,
    )

    @staticmethod
    def transform_hit_box_points(
        points: Point2List,
    ) -> Point2List:
        """
        Transform hit box points by rotating them 90 degrees clockwise.
        """
        return tuple(rotate_point(point[0], point[1], 0, 0, 90) for point in points)


class Rotate180Transform(Transform):
    """
    Rotate 180 degrees clockwise.
    """

    order = (
        VertexOrder.LOWER_RIGHT.value,
        VertexOrder.LOWER_LEFT.value,
        VertexOrder.UPPER_RIGHT.value,
        VertexOrder.UPPER_LEFT.value,
    )

    @staticmethod
    def transform_hit_box_points(
        points: Point2List,
    ) -> Point2List:
        """
        Transform hit box points by rotating them 180 degrees clockwise.
        """
        return tuple(rotate_point(point[0], point[1], 0, 0, 180) for point in points)


class Rotate270Transform(Transform):
    """
    Rotate 270 degrees clockwise.
    """

    order = (
        VertexOrder.UPPER_RIGHT.value,
        VertexOrder.LOWER_RIGHT.value,
        VertexOrder.UPPER_LEFT.value,
        VertexOrder.LOWER_LEFT.value,
    )

    @staticmethod
    def transform_hit_box_points(
        points: Point2List,
    ) -> Point2List:
        """
        Transform hit box points by rotating them 270 degrees clockwise.
        """
        return tuple(rotate_point(point[0], point[1], 0, 0, 270) for point in points)


class FlipLeftRightTransform(Transform):
    """
    Flip texture horizontally / left to right.
    """

    order = (
        VertexOrder.UPPER_RIGHT.value,
        VertexOrder.UPPER_LEFT.value,
        VertexOrder.LOWER_RIGHT.value,
        VertexOrder.LOWER_LEFT.value,
    )

    @staticmethod
    def transform_hit_box_points(
        points: Point2List,
    ) -> Point2List:
        """
        Transform hit box points by flipping them left to right.
        """
        return tuple((-point[0], point[1]) for point in points)


class FlipTopBottomTransform(Transform):
    """
    Flip texture vertically / top to bottom.
    """

    order = (
        VertexOrder.LOWER_LEFT.value,
        VertexOrder.LOWER_RIGHT.value,
        VertexOrder.UPPER_LEFT.value,
        VertexOrder.UPPER_RIGHT.value,
    )

    @staticmethod
    def transform_hit_box_points(
        points: Point2List,
    ) -> Point2List:
        """
        Transform hit box points by flipping them top to bottom.
        """
        return tuple((point[0], -point[1]) for point in points)


class TransposeTransform(Transform):
    """
    Transpose texture.
    """

    order = (
        VertexOrder.UPPER_LEFT.value,
        VertexOrder.LOWER_LEFT.value,
        VertexOrder.UPPER_RIGHT.value,
        VertexOrder.LOWER_RIGHT.value,
    )

    @staticmethod
    def transform_hit_box_points(
        points: Point2List,
    ) -> Point2List:
        """
        Transform hit box points by transposing them.
        """
        points = FlipLeftRightTransform.transform_hit_box_points(points)
        points = Rotate270Transform.transform_hit_box_points(points)
        return points


class TransverseTransform(Transform):
    """
    Transverse texture.
    """

    order = (
        VertexOrder.LOWER_RIGHT.value,
        VertexOrder.UPPER_RIGHT.value,
        VertexOrder.LOWER_LEFT.value,
        VertexOrder.UPPER_LEFT.value,
    )

    @staticmethod
    def transform_hit_box_points(
        points: Point2List,
    ) -> Point2List:
        """
        Transform hit box points by transversing them.
        """
        points = FlipLeftRightTransform.transform_hit_box_points(points)
        points = Rotate90Transform.transform_hit_box_points(points)
        return points


# Pre-calculated orientations. This can be calculated at runtime,
# but it's faster to just pre-calculate it.
# Key is the vertex order
# Value is the orientation (flip_left_right, flip_top_down, rotation)
ORIENTATIONS: dict[tuple[int, int, int, int], tuple[int, bool, bool]] = {
    (0, 1, 2, 3): (0, False, False),  # Default
    (2, 0, 3, 1): (90, False, False),  # Rotate 90
    (3, 2, 1, 0): (180, False, False),  # Rotate 180
    (1, 3, 0, 2): (270, False, False),  # Rotate 270
    (1, 0, 3, 2): (0, True, False),  # Flip left to right
    (2, 3, 0, 1): (0, False, True),  # Flip top to bottom
    (0, 2, 1, 3): (-90, True, False),  # Transpose
    (3, 1, 2, 0): (90, True, False),  # Transverse
}



"""
Platformer Game

python -m arcade.examples.platform_tutorial.01_open_window
"""
import arcade

# Constants
WINDOW_WIDTH = 1280
WINDOW_HEIGHT = 720
WINDOW_TITLE = "Platformer"


class GameView(arcade.Window):
    """
    Main application class.
    """

    def __init__(self):

        # Call the parent class to set up the window
        super().__init__(WINDOW_WIDTH, WINDOW_HEIGHT, WINDOW_TITLE)

        self.background_color = arcade.csscolor.CORNFLOWER_BLUE

    def setup(self):
        """Set up the game here. Call this function to restart the game."""
        pass

    def on_draw(self):
        """Render the screen."""

        # The clear method should always be called at the start of on_draw.
        # It clears the whole screen to whatever the background color is
        # set to. This ensures that you have a clean slate for drawing each
        # frame of the game.
        self.clear()

        # Code to draw other things will go here


def main():
    """Main function"""
    window = GameView()
    window.setup()
    arcade.run()


if __name__ == "__main__":
    main()



"""
Metadata about where an image is located in the atlas.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

from .base import TexCoords, TextureAtlasBase

if TYPE_CHECKING:
    from arcade.texture import ImageData


class AtlasRegion:
    """
    Stores information about where a texture is located.

    The texture coordinates are stored as a tuple of 8 floats
    (4 points, 2 floats each) in the following order:
    upper_left, upper_right, lower_left, lower_right.

    Layout::

        (0, 1)                                 (1, 1)
        +--------------------------------------+
        |          Atlas Texture               |
        |                                      |
        | (2)               (3)                |
        +-----------------+                    |
        |   Image         |                    |
        |                 |                    |
        |                 |                    |
        |                 |                    |
        |                 |                    |
        | (0)             | (1)                |
        +-----------------+--------------------+
        (0, 0)                                 (1, 0)

    Args:
        atlas:
            The atlas this region belongs to
        x:
            The x position of the texture
        y:
            The y position of the texture
        width:
            The width of the texture in pixels
        height:
            The height of the texture in pixels
        texture_coordinates:
            The texture coordinates for this region.
            If not provided, they will be calculated.
    """

    __slots__ = (
        "x",
        "y",
        "width",
        "height",
        "texture_coordinates",
    )

    def __init__(
        self,
        atlas: TextureAtlasBase,
        x: int,
        y: int,
        width: int,
        height: int,
        texture_coordinates: TexCoords | None = None,
    ):
        #: X position of the texture in the atlas
        self.x = x
        #: Y position of the texture in the atlas
        self.y = y
        #: Width of the texture in pixels
        self.width = width
        #: Height of the texture in pixels
        self.height = height
        # Calculate texture coordinates if not provided
        if texture_coordinates:
            #: The calculated texture coordinates for this region
            self.texture_coordinates = texture_coordinates
        else:
            # start_x, start_y, width, height
            # Width and height
            _width = self.width / atlas.width
            _height = self.height / atlas.height

            # Upper left corner. Note that are mapping the texture upside down and corners
            # are named as from the vertex position point of view.
            ul_x, ul_y = self.x / atlas.width, self.y / atlas.height

            # upper_left, upper_right, lower_left, lower_right
            self.texture_coordinates = (
                # upper_left
                ul_x,
                ul_y,
                # upper_right
                ul_x + _width,
                ul_y,
                # lower_left
                ul_x,
                ul_y + _height,
                # lower_right
                ul_x + _width,
                ul_y + _height,
            )

    def verify_image_size(self, image_data: ImageData):
        """
        Verify the image size.

        The internal image of a texture can potentially be tampered with
        at any point causing an atlas update to fail.

        Args:
            image_data: The image data to verify
        """
        if image_data.size != (self.width, self.height):
            raise RuntimeError(
                f"ImageData '{image_data.hash}' change their internal image "
                f"size from {self.width}x{self.height} to "
                f"{image_data.width}x{image_data.height}. "
                "It's not possible to fit this into the old allocated area in the atlas. "
            )

    def __repr__(self) -> str:
        return (
            f"<AtlasRegion x={self.x} y={self.y} width={self.width} "
            f"height={self.height} uvs={self.texture_coordinates}>"
        )



"""
Array Backed Grid Shown By Sprites

Show how to use a two-dimensional list/array to back the display of a
grid on-screen.

This version makes a grid of sprites instead of numbers. Instead of
iterating all the cells when the grid changes we simply just
swap the color of the selected sprite. This means this version
can handle very large grids and still have the same performance.

If Python and Arcade are installed, this example can be run from the command line with:
python -m arcade.examples.array_backed_grid_sprites_2
"""
import arcade

# Set how many rows and columns we will have
ROW_COUNT = 15
COLUMN_COUNT = 15

# This sets the WIDTH and HEIGHT of each grid location
WIDTH = 30
HEIGHT = 30

# This sets the margin between each cell
# and on the edges of the screen.
MARGIN = 5

# Do the math to figure out our screen dimensions
WINDOW_WIDTH = (WIDTH + MARGIN) * COLUMN_COUNT + MARGIN
WINDOW_HEIGHT = (HEIGHT + MARGIN) * ROW_COUNT + MARGIN
WINDOW_TITLE = "Array Backed Grid Buffered Example"


class GameView(arcade.View):
    """
    Main application class.
    """

    def __init__(self):
        """
        Set up the application.
        """
        super().__init__()

        # Set the background color of the window
        self.background_color = arcade.color.BLACK

        # One dimensional list of all sprites in the two-dimensional sprite list
        self.grid_sprite_list = arcade.SpriteList()

        # This will be a two-dimensional grid of sprites to mirror the two
        # dimensional grid of numbers. This points to the SAME sprites that are
        # in grid_sprite_list, just in a 2d manner.
        self.grid_sprites = []

        # Create a list of solid-color sprites to represent each grid location
        for row in range(ROW_COUNT):
            self.grid_sprites.append([])
            for column in range(COLUMN_COUNT):
                x = column * (WIDTH + MARGIN) + (WIDTH / 2 + MARGIN)
                y = row * (HEIGHT + MARGIN) + (HEIGHT / 2 + MARGIN)
                sprite = arcade.SpriteSolidColor(WIDTH, HEIGHT, color=arcade.color.WHITE)
                sprite.center_x = x
                sprite.center_y = y
                self.grid_sprite_list.append(sprite)
                self.grid_sprites[row].append(sprite)

    def on_draw(self):
        """
        Render the screen.
        """
        # We should always start by clearing the window pixels
        self.clear()

        # Batch draw the grid sprites
        self.grid_sprite_list.draw()

    def on_mouse_press(self, x, y, button, modifiers):
        """
        Called when the user presses a mouse button.
        """

        # Convert the clicked mouse position into grid coordinates
        column = int(x // (WIDTH + MARGIN))
        row = int(y // (HEIGHT + MARGIN))

        print(f"Click coordinates: ({x}, {y}). Grid coordinates: ({row}, {column})")

        # Make sure we are on-grid. It is possible to click in the upper right
        # corner in the margin and go to a grid location that doesn't exist
        if row >= ROW_COUNT or column >= COLUMN_COUNT:
            # Simply return from this method since nothing needs updating
            return

        # Flip the color of the sprite
        if self.grid_sprites[row][column].color == arcade.color.WHITE:
            self.grid_sprites[row][column].color = arcade.color.GREEN
        else:
            self.grid_sprites[row][column].color = arcade.color.WHITE


def main():
    """ Main function """
    # Create a window class. This is what actually shows up on screen
    window = arcade.Window(WINDOW_WIDTH, WINDOW_HEIGHT, WINDOW_TITLE)

    # Create the GameView
    game = GameView()

    # Show GameView on screen
    window.show_view(game)

    # Start the arcade game loop
    arcade.run()


if __name__ == "__main__":
    main()
